# Copyright 2017 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

"""Parse and convert amc motion capture data."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import collections

from dm_control.mujoco.wrapper import mjbindings
import numpy as np
from scipy import interpolate
from six.moves import range

mjlib = mjbindings.mjlib

MOCAP_DT = 1.0 / 120.0
CONVERSION_LENGTH = 0.056444

_CMU_MOCAP_JOINT_ORDER = (
    "root0",
    "root1",
    "root2",
    "root3",
    "root4",
    "root5",
    "lowerbackrx",
    "lowerbackry",
    "lowerbackrz",
    "upperbackrx",
    "upperbackry",
    "upperbackrz",
    "thoraxrx",
    "thoraxry",
    "thoraxrz",
    "lowerneckrx",
    "lowerneckry",
    "lowerneckrz",
    "upperneckrx",
    "upperneckry",
    "upperneckrz",
    "headrx",
    "headry",
    "headrz",
    "rclaviclery",
    "rclaviclerz",
    "rhumerusrx",
    "rhumerusry",
    "rhumerusrz",
    "rradiusrx",
    "rwristry",
    "rhandrx",
    "rhandrz",
    "rfingersrx",
    "rthumbrx",
    "rthumbrz",
    "lclaviclery",
    "lclaviclerz",
    "lhumerusrx",
    "lhumerusry",
    "lhumerusrz",
    "lradiusrx",
    "lwristry",
    "lhandrx",
    "lhandrz",
    "lfingersrx",
    "lthumbrx",
    "lthumbrz",
    "rfemurrx",
    "rfemurry",
    "rfemurrz",
    "rtibiarx",
    "rfootrx",
    "rfootrz",
    "rtoesrx",
    "lfemurrx",
    "lfemurry",
    "lfemurrz",
    "ltibiarx",
    "lfootrx",
    "lfootrz",
    "ltoesrx",
)

Converted = collections.namedtuple("Converted", ["qpos", "qvel", "time"])


def convert(file_name, physics, timestep):
    """Converts the parsed .amc values into qpos and qvel values and resamples.

    Args:
      file_name: The .amc file to be parsed and converted.
      physics: The corresponding physics instance.
      timestep: Desired output interval between resampled frames.

    Returns:
      A namedtuple with fields:
          `qpos`, a numpy array containing converted positional variables.
          `qvel`, a numpy array containing converted velocity variables.
          `time`, a numpy array containing the corresponding times.
    """
    frame_values = parse(file_name)
    joint2index = {}
    for name in physics.named.data.qpos.axes.row.names:
        joint2index[name] = physics.named.data.qpos.axes.row.convert_key_item(name)
    index2joint = {}
    for joint, index in joint2index.items():
        if isinstance(index, slice):
            indices = range(index.start, index.stop)
        else:
            indices = [index]
        for ii in indices:
            index2joint[ii] = joint

    # Convert frame_values to qpos
    amcvals2qpos_transformer = Amcvals2qpos(index2joint, _CMU_MOCAP_JOINT_ORDER)
    qpos_values = []
    for frame_value in frame_values:
        qpos_values.append(amcvals2qpos_transformer(frame_value))
    qpos_values = np.stack(qpos_values)  # Time by nq

    # Interpolate/resample.
    # Note: interpolate quaternions rather than euler angles (slerp).
    # see https://en.wikipedia.org/wiki/Slerp
    qpos_values_resampled = []
    time_vals = np.arange(0, len(frame_values) * MOCAP_DT - 1e-8, MOCAP_DT)
    time_vals_new = np.arange(0, len(frame_values) * MOCAP_DT, timestep)
    while time_vals_new[-1] > time_vals[-1]:
        time_vals_new = time_vals_new[:-1]

    for i in range(qpos_values.shape[1]):
        f = interpolate.splrep(time_vals, qpos_values[:, i])
        qpos_values_resampled.append(interpolate.splev(time_vals_new, f))

    qpos_values_resampled = np.stack(qpos_values_resampled)  # nq by ntime

    qvel_list = []
    for t in range(qpos_values_resampled.shape[1] - 1):
        p_tp1 = qpos_values_resampled[:, t + 1]
        p_t = qpos_values_resampled[:, t]
        qvel = [
            (p_tp1[:3] - p_t[:3]) / timestep,
            mj_quat2vel(mj_quatdiff(p_t[3:7], p_tp1[3:7]), timestep),
            (p_tp1[7:] - p_t[7:]) / timestep,
        ]
        qvel_list.append(np.concatenate(qvel))

    qvel_values_resampled = np.vstack(qvel_list).T

    return Converted(qpos_values_resampled, qvel_values_resampled, time_vals_new)


def parse(file_name):
    """Parses the amc file format."""
    values = []
    fid = open(file_name, "r")
    line = fid.readline().strip()
    frame_ind = 1
    first_frame = True
    while True:
        # Parse first frame.
        if first_frame and line[0] == str(frame_ind):
            first_frame = False
            frame_ind += 1
            frame_vals = []
            while True:
                line = fid.readline().strip()
                if not line or line == str(frame_ind):
                    values.append(np.array(frame_vals, dtype=np.float))
                    break
                tokens = line.split()
                frame_vals.extend(tokens[1:])
        # Parse other frames.
        elif line == str(frame_ind):
            frame_ind += 1
            frame_vals = []
            while True:
                line = fid.readline().strip()
                if not line or line == str(frame_ind):
                    values.append(np.array(frame_vals, dtype=np.float))
                    break
                tokens = line.split()
                frame_vals.extend(tokens[1:])
        else:
            line = fid.readline().strip()
            if not line:
                break
    return values


class Amcvals2qpos(object):
    """Callable that converts .amc values for a frame and to MuJoCo qpos format."""

    def __init__(self, index2joint, joint_order):
        """Initializes a new Amcvals2qpos instance.

        Args:
          index2joint: List of joint angles in .amc file.
          joint_order: List of joint names in MuJoco MJCF.
        """
        # Root is x,y,z, then quat.
        # need to get indices of qpos that order for amc default order
        self.qpos_root_xyz_ind = [0, 1, 2]
        self.root_xyz_ransform = (
            np.array([[1, 0, 0], [0, 0, -1], [0, 1, 0]]) * CONVERSION_LENGTH
        )
        self.qpos_root_quat_ind = [3, 4, 5, 6]
        amc2qpos_transform = np.zeros((len(index2joint), len(joint_order)))
        for i in range(len(index2joint)):
            for j in range(len(joint_order)):
                if index2joint[i] == joint_order[j]:
                    if "rx" in index2joint[i]:
                        amc2qpos_transform[i][j] = 1
                    elif "ry" in index2joint[i]:
                        amc2qpos_transform[i][j] = 1
                    elif "rz" in index2joint[i]:
                        amc2qpos_transform[i][j] = 1
        self.amc2qpos_transform = amc2qpos_transform

    def __call__(self, amc_val):
        """Converts a `.amc` frame to MuJoCo qpos format."""
        amc_val_rad = np.deg2rad(amc_val)
        qpos = np.dot(self.amc2qpos_transform, amc_val_rad)

        # Root.
        qpos[:3] = np.dot(self.root_xyz_ransform, amc_val[:3])
        qpos_quat = euler2quat(amc_val[3], amc_val[4], amc_val[5])
        qpos_quat = mj_quatprod(euler2quat(90, 0, 0), qpos_quat)

        for i, ind in enumerate(self.qpos_root_quat_ind):
            qpos[ind] = qpos_quat[i]

        return qpos


def euler2quat(ax, ay, az):
    """Converts euler angles to a quaternion.

    Note: rotation order is zyx

    Args:
      ax: Roll angle (deg)
      ay: Pitch angle (deg).
      az: Yaw angle (deg).

    Returns:
      A numpy array representing the rotation as a quaternion.
    """
    r1 = az
    r2 = ay
    r3 = ax

    c1 = np.cos(np.deg2rad(r1 / 2))
    s1 = np.sin(np.deg2rad(r1 / 2))
    c2 = np.cos(np.deg2rad(r2 / 2))
    s2 = np.sin(np.deg2rad(r2 / 2))
    c3 = np.cos(np.deg2rad(r3 / 2))
    s3 = np.sin(np.deg2rad(r3 / 2))

    q0 = c1 * c2 * c3 + s1 * s2 * s3
    q1 = c1 * c2 * s3 - s1 * s2 * c3
    q2 = c1 * s2 * c3 + s1 * c2 * s3
    q3 = s1 * c2 * c3 - c1 * s2 * s3

    return np.array([q0, q1, q2, q3])


def mj_quatprod(q, r):
    quaternion = np.zeros(4)
    mjlib.mju_mulQuat(quaternion, np.ascontiguousarray(q), np.ascontiguousarray(r))
    return quaternion


def mj_quat2vel(q, dt):
    vel = np.zeros(3)
    mjlib.mju_quat2Vel(vel, np.ascontiguousarray(q), dt)
    return vel


def mj_quatneg(q):
    quaternion = np.zeros(4)
    mjlib.mju_negQuat(quaternion, np.ascontiguousarray(q))
    return quaternion


def mj_quatdiff(source, target):
    return mj_quatprod(mj_quatneg(source), np.ascontiguousarray(target))



# Copyright 2019 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

"""Quadruped Domain."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import collections

from dm_control import mujoco
from dm_control.mujoco.wrapper import mjbindings
from dm_control.rl import control
from . import base
from . import common
from dm_control.utils import containers
from dm_control.utils import rewards
from dm_control.utils import xml_tools

from lxml import etree
import numpy as np
from scipy import ndimage

enums = mjbindings.enums
mjlib = mjbindings.mjlib


_DEFAULT_TIME_LIMIT = 20
_CONTROL_TIMESTEP = 0.02

# Horizontal speeds above which the move reward is 1.
_RUN_SPEED = 5
_WALK_SPEED = 0.5

# Constants related to terrain generation.
_HEIGHTFIELD_ID = 0
_TERRAIN_SMOOTHNESS = 0.15  # 0.0: maximally bumpy; 1.0: completely smooth.
_TERRAIN_BUMP_SCALE = 2  # Spatial scale of terrain bumps (in meters).

# Named model elements.
_TOES = ["toe_front_left", "toe_back_left", "toe_back_right", "toe_front_right"]
_WALLS = ["wall_px", "wall_py", "wall_nx", "wall_ny"]

SUITE = containers.TaggedTasks()


def make_model(
    floor_size=None, terrain=False, rangefinders=False, walls_and_ball=False
):
    """Returns the model XML string."""
    xml_string = common.read_model("quadruped.xml")
    parser = etree.XMLParser(remove_blank_text=True)
    mjcf = etree.XML(xml_string, parser)

    # Set floor size.
    if floor_size is not None:
        floor_geom = mjcf.find(".//geom[@name={!r}]".format("floor"))
        floor_geom.attrib["size"] = "{} {} .5".format(floor_size, floor_size)

    # Remove walls, ball and target.
    if not walls_and_ball:
        for wall in _WALLS:
            wall_geom = xml_tools.find_element(mjcf, "geom", wall)
            wall_geom.getparent().remove(wall_geom)

        # Remove ball.
        ball_body = xml_tools.find_element(mjcf, "body", "ball")
        ball_body.getparent().remove(ball_body)

        # Remove target.
        target_site = xml_tools.find_element(mjcf, "site", "target")
        target_site.getparent().remove(target_site)

    # Remove terrain.
    if not terrain:
        terrain_geom = xml_tools.find_element(mjcf, "geom", "terrain")
        terrain_geom.getparent().remove(terrain_geom)

    # Remove rangefinders if they're not used, as range computations can be
    # expensive, especially in a scene with heightfields.
    if not rangefinders:
        rangefinder_sensors = mjcf.findall(".//rangefinder")
        for rf in rangefinder_sensors:
            rf.getparent().remove(rf)

    return etree.tostring(mjcf, pretty_print=True)


@SUITE.add()
def walk(time_limit=_DEFAULT_TIME_LIMIT, random=None, environment_kwargs=None):
    """Returns the Walk task."""
    xml_string = make_model(floor_size=_DEFAULT_TIME_LIMIT * _WALK_SPEED)
    physics = Physics.from_xml_string(xml_string, common.ASSETS)
    task = Move(desired_speed=_WALK_SPEED, random=random)
    environment_kwargs = environment_kwargs or {}
    return control.Environment(
        physics,
        task,
        time_limit=time_limit,
        control_timestep=_CONTROL_TIMESTEP,
        **environment_kwargs
    )


@SUITE.add()
def run(time_limit=_DEFAULT_TIME_LIMIT, random=None, environment_kwargs=None):
    """Returns the Run task."""
    xml_string = make_model(floor_size=_DEFAULT_TIME_LIMIT * _RUN_SPEED)
    physics = Physics.from_xml_string(xml_string, common.ASSETS)
    task = Move(desired_speed=_RUN_SPEED, random=random)
    environment_kwargs = environment_kwargs or {}
    return control.Environment(
        physics,
        task,
        time_limit=time_limit,
        control_timestep=_CONTROL_TIMESTEP,
        **environment_kwargs
    )


@SUITE.add()
def escape(time_limit=_DEFAULT_TIME_LIMIT, random=None, environment_kwargs=None):
    """Returns the Escape task."""
    xml_string = make_model(floor_size=40, terrain=True, rangefinders=True)
    physics = Physics.from_xml_string(xml_string, common.ASSETS)
    task = Escape(random=random)
    environment_kwargs = environment_kwargs or {}
    return control.Environment(
        physics,
        task,
        time_limit=time_limit,
        control_timestep=_CONTROL_TIMESTEP,
        **environment_kwargs
    )


@SUITE.add()
def fetch(time_limit=_DEFAULT_TIME_LIMIT, random=None, environment_kwargs=None):
    """Returns the Fetch task."""
    xml_string = make_model(walls_and_ball=True)
    physics = Physics.from_xml_string(xml_string, common.ASSETS)
    task = Fetch(random=random)
    environment_kwargs = environment_kwargs or {}
    return control.Environment(
        physics,
        task,
        time_limit=time_limit,
        control_timestep=_CONTROL_TIMESTEP,
        **environment_kwargs
    )


class Physics(mujoco.Physics):
    """Physics simulation with additional features for the Quadruped domain."""

    def _reload_from_data(self, data):
        super(Physics, self)._reload_from_data(data)
        # Clear cached sensor names when the physics is reloaded.
        self._sensor_types_to_names = {}
        self._hinge_names = []

    def _get_sensor_names(self, *sensor_types):
        try:
            sensor_names = self._sensor_types_to_names[sensor_types]
        except KeyError:
            [sensor_ids] = np.where(np.in1d(self.model.sensor_type, sensor_types))
            sensor_names = [self.model.id2name(s_id, "sensor") for s_id in sensor_ids]
            self._sensor_types_to_names[sensor_types] = sensor_names
        return sensor_names

    def torso_upright(self):
        """Returns the dot-product of the torso z-axis and the global z-axis."""
        return np.asarray(self.named.data.xmat["torso", "zz"])

    def torso_velocity(self):
        """Returns the velocity of the torso, in the local frame."""
        return self.named.data.sensordata["velocimeter"].copy()

    def egocentric_state(self):
        """Returns the state without global orientation or position."""
        if not self._hinge_names:
            [hinge_ids] = np.nonzero(self.model.jnt_type == enums.mjtJoint.mjJNT_HINGE)
            self._hinge_names = [
                self.model.id2name(j_id, "joint") for j_id in hinge_ids
            ]
        return np.hstack(
            (
                self.named.data.qpos[self._hinge_names],
                self.named.data.qvel[self._hinge_names],
                self.data.act,
            )
        )

    def toe_positions(self):
        """Returns toe positions in egocentric frame."""
        torso_frame = self.named.data.xmat["torso"].reshape(3, 3)
        torso_pos = self.named.data.xpos["torso"]
        torso_to_toe = self.named.data.xpos[_TOES] - torso_pos
        return torso_to_toe.dot(torso_frame)

    def force_torque(self):
        """Returns scaled force/torque sensor readings at the toes."""
        force_torque_sensors = self._get_sensor_names(
            enums.mjtSensor.mjSENS_FORCE, enums.mjtSensor.mjSENS_TORQUE
        )
        return np.arcsinh(self.named.data.sensordata[force_torque_sensors])

    def imu(self):
        """Returns IMU-like sensor readings."""
        imu_sensors = self._get_sensor_names(
            enums.mjtSensor.mjSENS_GYRO, enums.mjtSensor.mjSENS_ACCELEROMETER
        )
        return self.named.data.sensordata[imu_sensors]

    def rangefinder(self):
        """Returns scaled rangefinder sensor readings."""
        rf_sensors = self._get_sensor_names(enums.mjtSensor.mjSENS_RANGEFINDER)
        rf_readings = self.named.data.sensordata[rf_sensors]
        no_intersection = -1.0
        return np.where(rf_readings == no_intersection, 1.0, np.tanh(rf_readings))

    def origin_distance(self):
        """Returns the distance from the origin to the workspace."""
        return np.asarray(np.linalg.norm(self.named.data.site_xpos["workspace"]))

    def origin(self):
        """Returns origin position in the torso frame."""
        torso_frame = self.named.data.xmat["torso"].reshape(3, 3)
        torso_pos = self.named.data.xpos["torso"]
        return -torso_pos.dot(torso_frame)

    def ball_state(self):
        """Returns ball position and velocity relative to the torso frame."""
        data = self.named.data
        torso_frame = data.xmat["torso"].reshape(3, 3)
        ball_rel_pos = data.xpos["ball"] - data.xpos["torso"]
        ball_rel_vel = data.qvel["ball_root"][:3] - data.qvel["root"][:3]
        ball_rot_vel = data.qvel["ball_root"][3:]
        ball_state = np.vstack((ball_rel_pos, ball_rel_vel, ball_rot_vel))
        return ball_state.dot(torso_frame).ravel()

    def target_position(self):
        """Returns target position in torso frame."""
        torso_frame = self.named.data.xmat["torso"].reshape(3, 3)
        torso_pos = self.named.data.xpos["torso"]
        torso_to_target = self.named.data.site_xpos["target"] - torso_pos
        return torso_to_target.dot(torso_frame)

    def ball_to_target_distance(self):
        """Returns horizontal distance from the ball to the target."""
        ball_to_target = (
            self.named.data.site_xpos["target"] - self.named.data.xpos["ball"]
        )
        return np.linalg.norm(ball_to_target[:2])

    def self_to_ball_distance(self):
        """Returns horizontal distance from the quadruped workspace to the ball."""
        self_to_ball = (
            self.named.data.site_xpos["workspace"] - self.named.data.xpos["ball"]
        )
        return np.linalg.norm(self_to_ball[:2])


def _find_non_contacting_height(physics, orientation, x_pos=0.0, y_pos=0.0):
    """Find a height with no contacts given a body orientation.

    Args:
      physics: An instance of `Physics`.
      orientation: A quaternion.
      x_pos: A float. Position along global x-axis.
      y_pos: A float. Position along global y-axis.
    Raises:
      RuntimeError: If a non-contacting configuration has not been found after
      10,000 attempts.
    """
    z_pos = 0.0  # Start embedded in the floor.
    num_contacts = 1
    num_attempts = 0
    # Move up in 1cm increments until no contacts.
    while num_contacts > 0:
        try:
            with physics.reset_context():
                physics.named.data.qpos["root"][:3] = x_pos, y_pos, z_pos
                physics.named.data.qpos["root"][3:] = orientation
        except control.PhysicsError:
            # We may encounter a PhysicsError here due to filling the contact
            # buffer, in which case we simply increment the height and continue.
            pass
        num_contacts = physics.data.ncon
        z_pos += 0.01
        num_attempts += 1
        if num_attempts > 10000:
            raise RuntimeError("Failed to find a non-contacting configuration.")


def _common_observations(physics):
    """Returns the observations common to all tasks."""
    obs = collections.OrderedDict()
    obs["egocentric_state"] = physics.egocentric_state()
    obs["torso_velocity"] = physics.torso_velocity()
    obs["torso_upright"] = physics.torso_upright()
    obs["imu"] = physics.imu()
    obs["force_torque"] = physics.force_torque()
    return obs


def _upright_reward(physics, deviation_angle=0):
    """Returns a reward proportional to how upright the torso is.

    Args:
      physics: an instance of `Physics`.
      deviation_angle: A float, in degrees. The reward is 0 when the torso is
        exactly upside-down and 1 when the torso's z-axis is less than
        `deviation_angle` away from the global z-axis.
    """
    deviation = np.cos(np.deg2rad(deviation_angle))
    return rewards.tolerance(
        physics.torso_upright(),
        bounds=(deviation, float("inf")),
        sigmoid="linear",
        margin=1 + deviation,
        value_at_margin=0,
    )


class Move(base.Task):
    """A quadruped task solved by moving forward at a designated speed."""

    def __init__(self, desired_speed, random=None):
        """Initializes an instance of `Move`.

        Args:
          desired_speed: A float. If this value is zero, reward is given simply
            for standing upright. Otherwise this specifies the horizontal velocity
            at which the velocity-dependent reward component is maximized.
          random: Optional, either a `numpy.random.RandomState` instance, an
            integer seed for creating a new `RandomState`, or None to select a seed
            automatically (default).
        """
        self._desired_speed = desired_speed
        super(Move, self).__init__(random=random)

    def initialize_episode(self, physics):
        """Sets the state of the environment at the start of each episode.

        Args:
          physics: An instance of `Physics`.

        """
        # Initial configuration.
        orientation = self.random.randn(4)
        orientation /= np.linalg.norm(orientation)
        _find_non_contacting_height(physics, orientation)
        super(Move, self).initialize_episode(physics)

    def get_observation(self, physics):
        """Returns an observation to the agent."""
        return _common_observations(physics)

    def get_reward(self, physics):
        """Returns a reward to the agent."""

        # Move reward term.
        move_reward = rewards.tolerance(
            physics.torso_velocity()[0],
            bounds=(self._desired_speed, float("inf")),
            margin=self._desired_speed,
            value_at_margin=0.5,
            sigmoid="linear",
        )

        return _upright_reward(physics) * move_reward


class Escape(base.Task):
    """A quadruped task solved by escaping a bowl-shaped terrain."""

    def initialize_episode(self, physics):
        """Sets the state of the environment at the start of each episode.

        Args:
          physics: An instance of `Physics`.

        """
        # Get heightfield resolution, assert that it is square.
        res = physics.model.hfield_nrow[_HEIGHTFIELD_ID]
        assert res == physics.model.hfield_ncol[_HEIGHTFIELD_ID]
        # Sinusoidal bowl shape.
        row_grid, col_grid = np.ogrid[-1 : 1 : res * 1j, -1 : 1 : res * 1j]
        radius = np.clip(np.sqrt(col_grid ** 2 + row_grid ** 2), 0.04, 1)
        bowl_shape = 0.5 - np.cos(2 * np.pi * radius) / 2
        # Random smooth bumps.
        terrain_size = 2 * physics.model.hfield_size[_HEIGHTFIELD_ID, 0]
        bump_res = int(terrain_size / _TERRAIN_BUMP_SCALE)
        bumps = self.random.uniform(_TERRAIN_SMOOTHNESS, 1, (bump_res, bump_res))
        smooth_bumps = ndimage.zoom(bumps, res / float(bump_res))
        # Terrain is elementwise product.
        terrain = bowl_shape * smooth_bumps
        start_idx = physics.model.hfield_adr[_HEIGHTFIELD_ID]
        physics.model.hfield_data[start_idx : start_idx + res ** 2] = terrain.ravel()
        super(Escape, self).initialize_episode(physics)

        # If we have a rendering context, we need to re-upload the modified
        # heightfield data.
        if physics.contexts:
            with physics.contexts.gl.make_current() as ctx:
                ctx.call(
                    mjlib.mjr_uploadHField,
                    physics.model.ptr,
                    physics.contexts.mujoco.ptr,
                    _HEIGHTFIELD_ID,
                )

        # Initial configuration.
        orientation = self.random.randn(4)
        orientation /= np.linalg.norm(orientation)
        _find_non_contacting_height(physics, orientation)

    def get_observation(self, physics):
        """Returns an observation to the agent."""
        obs = _common_observations(physics)
        obs["origin"] = physics.origin()
        obs["rangefinder"] = physics.rangefinder()
        return obs

    def get_reward(self, physics):
        """Returns a reward to the agent."""

        # Escape reward term.
        terrain_size = physics.model.hfield_size[_HEIGHTFIELD_ID, 0]
        escape_reward = rewards.tolerance(
            physics.origin_distance(),
            bounds=(terrain_size, float("inf")),
            margin=terrain_size,
            value_at_margin=0,
            sigmoid="linear",
        )

        return _upright_reward(physics, deviation_angle=20) * escape_reward


class Fetch(base.Task):
    """A quadruped task solved by bringing a ball to the origin."""

    def initialize_episode(self, physics):
        """Sets the state of the environment at the start of each episode.

        Args:
          physics: An instance of `Physics`.

        """
        # Initial configuration, random azimuth and horizontal position.
        azimuth = self.random.uniform(0, 2 * np.pi)
        orientation = np.array((np.cos(azimuth / 2), 0, 0, np.sin(azimuth / 2)))
        spawn_radius = 0.9 * physics.named.model.geom_size["floor", 0]
        x_pos, y_pos = self.random.uniform(-spawn_radius, spawn_radius, size=(2,))
        _find_non_contacting_height(physics, orientation, x_pos, y_pos)

        # Initial ball state.
        physics.named.data.qpos["ball_root"][:2] = self.random.uniform(
            -spawn_radius, spawn_radius, size=(2,)
        )
        physics.named.data.qpos["ball_root"][2] = 2
        physics.named.data.qvel["ball_root"][:2] = 5 * self.random.randn(2)
        super(Fetch, self).initialize_episode(physics)

    def get_observation(self, physics):
        """Returns an observation to the agent."""
        obs = _common_observations(physics)
        obs["ball_state"] = physics.ball_state()
        obs["target_position"] = physics.target_position()
        return obs

    def get_reward(self, physics):
        """Returns a reward to the agent."""

        # Reward for moving close to the ball.
        arena_radius = physics.named.model.geom_size["floor", 0] * np.sqrt(2)
        workspace_radius = physics.named.model.site_size["workspace", 0]
        ball_radius = physics.named.model.geom_size["ball", 0]
        reach_reward = rewards.tolerance(
            physics.self_to_ball_distance(),
            bounds=(0, workspace_radius + ball_radius),
            sigmoid="linear",
            margin=arena_radius,
            value_at_margin=0,
        )

        # Reward for bringing the ball to the target.
        target_radius = physics.named.model.site_size["target", 0]
        fetch_reward = rewards.tolerance(
            physics.ball_to_target_distance(),
            bounds=(0, target_radius),
            sigmoid="linear",
            margin=arena_radius,
            value_at_margin=0,
        )

        reach_then_fetch = reach_reward * (0.5 + 0.5 * fetch_reward)

        return _upright_reward(physics) * reach_then_fetch



# Copyright 2017 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

"""Planar Stacker domain."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import collections

from dm_control import mujoco
from dm_control.rl import control
from . import base
from . import common
from dm_control.utils import containers
from dm_control.utils import rewards
from dm_control.utils import xml_tools

from lxml import etree
import numpy as np


_CLOSE = 0.01  # (Meters) Distance below which a thing is considered close.
_CONTROL_TIMESTEP = 0.01  # (Seconds)
_TIME_LIMIT = 10  # (Seconds)
_ARM_JOINTS = [
    "arm_root",
    "arm_shoulder",
    "arm_elbow",
    "arm_wrist",
    "finger",
    "fingertip",
    "thumb",
    "thumbtip",
]

SUITE = containers.TaggedTasks()


def make_model(n_boxes):
    """Returns a tuple containing the model XML string and a dict of assets."""
    xml_string = common.read_model("stacker.xml")
    parser = etree.XMLParser(remove_blank_text=True)
    mjcf = etree.XML(xml_string, parser)

    # Remove unused boxes
    for b in range(n_boxes, 4):
        box = xml_tools.find_element(mjcf, "body", "box" + str(b))
        box.getparent().remove(box)

    return etree.tostring(mjcf, pretty_print=True), common.ASSETS


@SUITE.add("hard")
def stack_2(
    fully_observable=True, time_limit=_TIME_LIMIT, random=None, environment_kwargs=None
):
    """Returns stacker task with 2 boxes."""
    n_boxes = 2
    physics = Physics.from_xml_string(*make_model(n_boxes=n_boxes))
    task = Stack(n_boxes=n_boxes, fully_observable=fully_observable, random=random)
    environment_kwargs = environment_kwargs or {}
    return control.Environment(
        physics,
        task,
        control_timestep=_CONTROL_TIMESTEP,
        time_limit=time_limit,
        **environment_kwargs
    )


@SUITE.add("hard")
def stack_4(
    fully_observable=True, time_limit=_TIME_LIMIT, random=None, environment_kwargs=None
):
    """Returns stacker task with 4 boxes."""
    n_boxes = 4
    physics = Physics.from_xml_string(*make_model(n_boxes=n_boxes))
    task = Stack(n_boxes=n_boxes, fully_observable=fully_observable, random=random)
    environment_kwargs = environment_kwargs or {}
    return control.Environment(
        physics,
        task,
        control_timestep=_CONTROL_TIMESTEP,
        time_limit=time_limit,
        **environment_kwargs
    )


class Physics(mujoco.Physics):
    """Physics with additional features for the Planar Manipulator domain."""

    def bounded_joint_pos(self, joint_names):
        """Returns joint positions as (sin, cos) values."""
        joint_pos = self.named.data.qpos[joint_names]
        return np.vstack([np.sin(joint_pos), np.cos(joint_pos)]).T

    def joint_vel(self, joint_names):
        """Returns joint velocities."""
        return self.named.data.qvel[joint_names]

    def body_2d_pose(self, body_names, orientation=True):
        """Returns positions and/or orientations of bodies."""
        if not isinstance(body_names, str):
            body_names = np.array(body_names).reshape(-1, 1)  # Broadcast indices.
        pos = self.named.data.xpos[body_names, ["x", "z"]]
        if orientation:
            ori = self.named.data.xquat[body_names, ["qw", "qy"]]
            return np.hstack([pos, ori])
        else:
            return pos

    def touch(self):
        return np.log1p(self.data.sensordata)

    def site_distance(self, site1, site2):
        site1_to_site2 = np.diff(self.named.data.site_xpos[[site2, site1]], axis=0)
        return np.linalg.norm(site1_to_site2)


class Stack(base.Task):
    """A Stack `Task`: stack the boxes."""

    def __init__(self, n_boxes, fully_observable, random=None):
        """Initialize an instance of the `Stack` task.

        Args:
          n_boxes: An `int`, number of boxes to stack.
          fully_observable: A `bool`, whether the observation should contain the
            positions and velocities of the boxes and the location of the target.
          random: Optional, either a `numpy.random.RandomState` instance, an
            integer seed for creating a new `RandomState`, or None to select a seed
            automatically (default).
        """
        self._n_boxes = n_boxes
        self._box_names = ["box" + str(b) for b in range(n_boxes)]
        self._box_joint_names = []
        for name in self._box_names:
            for dim in "xyz":
                self._box_joint_names.append("_".join([name, dim]))
        self._fully_observable = fully_observable
        super(Stack, self).__init__(random=random)

    def initialize_episode(self, physics):
        """Sets the state of the environment at the start of each episode."""
        # Local aliases
        randint = self.random.randint
        uniform = self.random.uniform
        model = physics.named.model
        data = physics.named.data

        # Find a collision-free random initial configuration.
        penetrating = True
        while penetrating:

            # Randomise angles of arm joints.
            is_limited = model.jnt_limited[_ARM_JOINTS].astype(np.bool)
            joint_range = model.jnt_range[_ARM_JOINTS]
            lower_limits = np.where(is_limited, joint_range[:, 0], -np.pi)
            upper_limits = np.where(is_limited, joint_range[:, 1], np.pi)
            angles = uniform(lower_limits, upper_limits)
            data.qpos[_ARM_JOINTS] = angles

            # Symmetrize hand.
            data.qpos["finger"] = data.qpos["thumb"]

            # Randomise target location.
            target_height = 2 * randint(self._n_boxes) + 1
            box_size = model.geom_size["target", 0]
            model.body_pos["target", "z"] = box_size * target_height
            model.body_pos["target", "x"] = uniform(-0.37, 0.37)

            # Randomise box locations.
            for name in self._box_names:
                data.qpos[name + "_x"] = uniform(0.1, 0.3)
                data.qpos[name + "_z"] = uniform(0, 0.7)
                data.qpos[name + "_y"] = uniform(0, 2 * np.pi)

            # Check for collisions.
            physics.after_reset()
            penetrating = physics.data.ncon > 0

        super(Stack, self).initialize_episode(physics)

    def get_observation(self, physics):
        """Returns either features or only sensors (to be used with pixels)."""
        obs = collections.OrderedDict()
        obs["arm_pos"] = physics.bounded_joint_pos(_ARM_JOINTS)
        obs["arm_vel"] = physics.joint_vel(_ARM_JOINTS)
        obs["touch"] = physics.touch()
        if self._fully_observable:
            obs["hand_pos"] = physics.body_2d_pose("hand")
            obs["box_pos"] = physics.body_2d_pose(self._box_names)
            obs["box_vel"] = physics.joint_vel(self._box_joint_names)
            obs["target_pos"] = physics.body_2d_pose("target", orientation=False)
        return obs

    def get_reward(self, physics):
        """Returns a reward to the agent."""
        box_size = physics.named.model.geom_size["target", 0]
        min_box_to_target_distance = min(
            physics.site_distance(name, "target") for name in self._box_names
        )
        box_is_close = rewards.tolerance(
            min_box_to_target_distance, margin=2 * box_size
        )
        hand_to_target_distance = physics.site_distance("grasp", "target")
        hand_is_far = rewards.tolerance(
            hand_to_target_distance, bounds=(0.1, float("inf")), margin=_CLOSE
        )
        return box_is_close * hand_is_far



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved
from typing import Any, Dict, Tuple, Union

import numpy as np

TaskObsType = Union[str, int, float, np.ndarray]
ActionType = Union[str, int, float, np.ndarray]
EnvObsType = Union[np.ndarray]
ObsType = Dict[str, Union[EnvObsType, TaskObsType]]
RewardType = float
DoneType = bool
InfoType = Dict[str, Any]
StepReturnType = Tuple[ObsType, RewardType, DoneType, InfoType]
EnvStepReturnType = Tuple[EnvObsType, RewardType, DoneType, InfoType]
TaskStateType = Any



# Copyright (c) Facebook, Inc. and its affiliates.
# All rights reserved.

from typing import Any, Dict, Optional

from gym import error
from gym.core import Env
from gym.envs.registration import EnvRegistry, EnvSpec


class MultitaskEnvSpec(EnvSpec):  # type: ignore[misc]
    def __init__(
        self,
        id: str,
        entry_point: Optional[str] = None,
        reward_threshold: Optional[int] = None,
        kwargs: Optional[Dict[str, Any]] = None,
        nondeterministic: bool = False,
        max_episode_steps: Optional[int] = None,
        test_kwargs: Optional[Dict[str, Any]] = None,
    ):
        """A specification for a particular instance of the environment.
        Used to register the parameters for official evaluations.

        Args:
            id (str): The official environment ID
            entry_point (Optional[str]): The Python entrypoint of the
                environment class (e.g. module.name:Class)
            reward_threshold (Optional[int]): The reward threshold before
                the task is considered solved
            kwargs (dict): The kwargs to pass to the environment class
            nondeterministic (bool): Whether this environment is
                non-deterministic even after seeding
            max_episode_steps (Optional[int]): The maximum number of steps
                that an episode can consist of
            test_kwargs (Optional[Dict[str, Any]], optional): Dictionary
                to specify parameters for automated testing. Defaults to
                None.

        """
        super().__init__(
            id=id,
            entry_point=entry_point,
            reward_threshold=reward_threshold,
            nondeterministic=nondeterministic,
            max_episode_steps=max_episode_steps,
            kwargs=kwargs,
        )
        self.test_kwargs = test_kwargs

    def __repr__(self) -> str:
        return f"MultitaskEnvSpec({self.id})"

    @property
    def kwargs(self) -> Dict[str, Any]:
        return self._kwargs  # type: ignore[no-any-return]


class MultiEnvRegistry(EnvRegistry):  # type: ignore[misc]
    def __init__(self) -> None:
        super().__init__()

    def register(self, id: str, **kwargs: Any) -> None:
        if id in self.env_specs:
            raise error.Error("Cannot re-register id: {}".format(id))
        self.env_specs[id] = MultitaskEnvSpec(id, **kwargs)


# Have a global registry
mtenv_registry = MultiEnvRegistry()


def register(id: str, **kwargs: Any) -> None:
    return mtenv_registry.register(id, **kwargs)


def make(id: str, **kwargs: Any) -> Env:
    env = mtenv_registry.make(id, **kwargs)
    assert isinstance(env, Env)
    return env


def spec(id: str) -> MultitaskEnvSpec:
    spec = mtenv_registry.spec(id)
    assert isinstance(spec, MultitaskEnvSpec)
    return spec



# Copyright 2017 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

"""Wrapper that adds pixel observations to a control environment."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import collections

import dm_env
from dm_env import specs

STATE_KEY = "state"


class Wrapper(dm_env.Environment):
    """Wraps a control environment and adds a rendered pixel observation."""

    def __init__(
        self, env, pixels_only=True, render_kwargs=None, observation_key="pixels"
    ):
        """Initializes a new pixel Wrapper.

        Args:
          env: The environment to wrap.
          pixels_only: If True (default), the original set of 'state' observations
            returned by the wrapped environment will be discarded, and the
            `OrderedDict` of observations will only contain pixels. If False, the
            `OrderedDict` will contain the original observations as well as the
            pixel observations.
          render_kwargs: Optional `dict` containing keyword arguments passed to the
            `mujoco.Physics.render` method.
          observation_key: Optional custom string specifying the pixel observation's
            key in the `OrderedDict` of observations. Defaults to 'pixels'.

        Raises:
          ValueError: If `env`'s observation spec is not compatible with the
            wrapper. Supported formats are a single array, or a dict of arrays.
          ValueError: If `env`'s observation already contains the specified
            `observation_key`.
        """
        if render_kwargs is None:
            render_kwargs = {}

        wrapped_observation_spec = env.observation_spec()

        if isinstance(wrapped_observation_spec, specs.Array):
            self._observation_is_dict = False
            invalid_keys = set([STATE_KEY])
        elif isinstance(wrapped_observation_spec, collections.MutableMapping):
            self._observation_is_dict = True
            invalid_keys = set(wrapped_observation_spec.keys())
        else:
            raise ValueError("Unsupported observation spec structure.")

        if not pixels_only and observation_key in invalid_keys:
            raise ValueError(
                "Duplicate or reserved observation key {!r}.".format(observation_key)
            )

        if pixels_only:
            self._observation_spec = collections.OrderedDict()
        elif self._observation_is_dict:
            self._observation_spec = wrapped_observation_spec.copy()
        else:
            self._observation_spec = collections.OrderedDict()
            self._observation_spec[STATE_KEY] = wrapped_observation_spec

        # Extend observation spec.
        pixels = env.physics.render(**render_kwargs)
        pixels_spec = specs.Array(
            shape=pixels.shape, dtype=pixels.dtype, name=observation_key
        )
        self._observation_spec[observation_key] = pixels_spec

        self._env = env
        self._pixels_only = pixels_only
        self._render_kwargs = render_kwargs
        self._observation_key = observation_key

    def reset(self):
        time_step = self._env.reset()
        return self._add_pixel_observation(time_step)

    def step(self, action):
        time_step = self._env.step(action)
        return self._add_pixel_observation(time_step)

    def observation_spec(self):
        return self._observation_spec

    def action_spec(self):
        return self._env.action_spec()

    def _add_pixel_observation(self, time_step):
        if self._pixels_only:
            observation = collections.OrderedDict()
        elif self._observation_is_dict:
            observation = type(time_step.observation)(time_step.observation)
        else:
            observation = collections.OrderedDict()
            observation[STATE_KEY] = time_step.observation

        pixels = self._env.physics.render(**self._render_kwargs)
        observation[self._observation_key] = pixels
        return time_step._replace(observation=observation)

    def __getattr__(self, name):
        return getattr(self._env, name)



# Copyright 2017 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

"""Cartpole domain."""

from __future__ import absolute_import, division, print_function

import collections

import numpy as np
from dm_control import mujoco
from dm_control.rl import control
from dm_control.utils import containers, rewards
from . import base, common
from lxml import etree
from six.moves import range

_DEFAULT_TIME_LIMIT = 10
SUITE = containers.TaggedTasks()


def get_model_and_assets(num_poles=1, xml_file_id=None):
    """Returns a tuple containing the model XML string and a dict of assets."""
    return _make_model(num_poles, xml_file_id), common.ASSETS


@SUITE.add("benchmarking")
def balance(
    time_limit=_DEFAULT_TIME_LIMIT, random=None, environment_kwargs=None,
):
    """Returns the Cartpole Balance task."""
    physics = Physics.from_xml_string(*get_model_and_assets())
    task = Balance(swing_up=False, sparse=False, random=random)
    environment_kwargs = environment_kwargs or {}
    return control.Environment(
        physics, task, time_limit=time_limit, **environment_kwargs
    )


@SUITE.add("benchmarking")
def balance_sparse(
    time_limit=_DEFAULT_TIME_LIMIT, random=None, environment_kwargs=None
):
    """Returns the sparse reward variant of the Cartpole Balance task."""
    physics = Physics.from_xml_string(*get_model_and_assets())
    task = Balance(swing_up=False, sparse=True, random=random)
    environment_kwargs = environment_kwargs or {}
    return control.Environment(
        physics, task, time_limit=time_limit, **environment_kwargs
    )


@SUITE.add("benchmarking")
def swingup(
    time_limit=_DEFAULT_TIME_LIMIT,
    xml_file_id=None,
    random=None,
    environment_kwargs=None,
):
    """Returns the Cartpole Swing-Up task."""
    physics = Physics.from_xml_string(*get_model_and_assets(xml_file_id=xml_file_id))
    task = Balance(swing_up=True, sparse=False, random=random)
    environment_kwargs = environment_kwargs or {}
    return control.Environment(
        physics, task, time_limit=time_limit, **environment_kwargs
    )


@SUITE.add("benchmarking")
def swingup_sparse(
    time_limit=_DEFAULT_TIME_LIMIT, random=None, environment_kwargs=None
):
    """Returns the sparse reward variant of teh Cartpole Swing-Up task."""
    physics = Physics.from_xml_string(*get_model_and_assets())
    task = Balance(swing_up=True, sparse=True, random=random)
    environment_kwargs = environment_kwargs or {}
    return control.Environment(
        physics, task, time_limit=time_limit, **environment_kwargs
    )


@SUITE.add()
def two_poles(time_limit=_DEFAULT_TIME_LIMIT, random=None, environment_kwargs=None):
    """Returns the Cartpole Balance task with two poles."""
    physics = Physics.from_xml_string(*get_model_and_assets(num_poles=2))
    task = Balance(swing_up=True, sparse=False, random=random)
    environment_kwargs = environment_kwargs or {}
    return control.Environment(
        physics, task, time_limit=time_limit, **environment_kwargs
    )


@SUITE.add()
def three_poles(
    time_limit=_DEFAULT_TIME_LIMIT,
    random=None,
    num_poles=3,
    sparse=False,
    environment_kwargs=None,
):
    """Returns the Cartpole Balance task with three or more poles."""
    physics = Physics.from_xml_string(*get_model_and_assets(num_poles=num_poles))
    task = Balance(swing_up=True, sparse=sparse, random=random)
    environment_kwargs = environment_kwargs or {}
    return control.Environment(
        physics, task, time_limit=time_limit, **environment_kwargs
    )


def _make_model(n_poles, xml_file_id=None):
    """Generates an xml string defining a cart with `n_poles` bodies."""
    if xml_file_id is not None:
        filename = f"cartpole_{xml_file_id}.xml"
        print(filename)
    else:
        filename = f"cartpole.xml"
    xml_string = common.read_model(filename)
    if n_poles == 1:
        return xml_string
    mjcf = etree.fromstring(xml_string)
    parent = mjcf.find("./worldbody/body/body")  # Find first pole.
    # Make chain of poles.
    for pole_index in range(2, n_poles + 1):
        child = etree.Element(
            "body", name="pole_{}".format(pole_index), pos="0 0 1", childclass="pole"
        )
        etree.SubElement(child, "joint", name="hinge_{}".format(pole_index))
        etree.SubElement(child, "geom", name="pole_{}".format(pole_index))
        parent.append(child)
        parent = child
    # Move plane down.
    floor = mjcf.find("./worldbody/geom")
    floor.set("pos", "0 0 {}".format(1 - n_poles - 0.05))
    # Move cameras back.
    cameras = mjcf.findall("./worldbody/camera")
    cameras[0].set("pos", "0 {} 1".format(-1 - 2 * n_poles))
    cameras[1].set("pos", "0 {} 2".format(-2 * n_poles))
    return etree.tostring(mjcf, pretty_print=True)


class Physics(mujoco.Physics):
    """Physics simulation with additional features for the Cartpole domain."""

    def cart_position(self):
        """Returns the position of the cart."""
        return self.named.data.qpos["slider"][0]

    def angular_vel(self):
        """Returns the angular velocity of the pole."""
        return self.data.qvel[1:]

    def pole_angle_cosine(self):
        """Returns the cosine of the pole angle."""
        return self.named.data.xmat[2:, "zz"]

    def bounded_position(self):
        """Returns the state, with pole angle split into sin/cos."""
        return np.hstack(
            (self.cart_position(), self.named.data.xmat[2:, ["zz", "xz"]].ravel())
        )


class Balance(base.Task):
    """A Cartpole `Task` to balance the pole.

    State is initialized either close to the target configuration or at a random
    configuration.
    """

    _CART_RANGE = (-0.25, 0.25)
    _ANGLE_COSINE_RANGE = (0.995, 1)

    def __init__(self, swing_up, sparse, random=None):
        """Initializes an instance of `Balance`.

        Args:
          swing_up: A `bool`, which if `True` sets the cart to the middle of the
            slider and the pole pointing towards the ground. Otherwise, sets the
            cart to a random position on the slider and the pole to a random
            near-vertical position.
          sparse: A `bool`, whether to return a sparse or a smooth reward.
          random: Optional, either a `numpy.random.RandomState` instance, an
            integer seed for creating a new `RandomState`, or None to select a seed
            automatically (default).
        """
        self._sparse = sparse
        self._swing_up = swing_up
        super(Balance, self).__init__(random=random)

    def initialize_episode(self, physics):
        """Sets the state of the environment at the start of each episode.

        Initializes the cart and pole according to `swing_up`, and in both cases
        adds a small random initial velocity to break symmetry.

        Args:
          physics: An instance of `Physics`.
        """
        nv = physics.model.nv
        if self._swing_up:
            physics.named.data.qpos["slider"] = 0.01 * self.random.randn()
            physics.named.data.qpos["hinge_1"] = np.pi + 0.01 * self.random.randn()
            physics.named.data.qpos[2:] = 0.1 * self.random.randn(nv - 2)
        else:
            physics.named.data.qpos["slider"] = self.random.uniform(-0.1, 0.1)
            physics.named.data.qpos[1:] = self.random.uniform(-0.034, 0.034, nv - 1)
        physics.named.data.qvel[:] = 0.01 * self.random.randn(physics.model.nv)
        super(Balance, self).initialize_episode(physics)

    def get_observation(self, physics):
        """Returns an observation of the (bounded) physics state."""
        obs = collections.OrderedDict()
        obs["position"] = physics.bounded_position()
        obs["velocity"] = physics.velocity()
        return obs

    def _get_reward(self, physics, sparse):
        if sparse:
            cart_in_bounds = rewards.tolerance(
                physics.cart_position(), self._CART_RANGE
            )
            angle_in_bounds = rewards.tolerance(
                physics.pole_angle_cosine(), self._ANGLE_COSINE_RANGE
            ).prod()
            return cart_in_bounds * angle_in_bounds
        else:
            upright = (physics.pole_angle_cosine() + 1) / 2
            centered = rewards.tolerance(physics.cart_position(), margin=2)
            centered = (1 + centered) / 2
            small_control = rewards.tolerance(
                physics.control(), margin=1, value_at_margin=0, sigmoid="quadratic"
            )[0]
            small_control = (4 + small_control) / 5
            small_velocity = rewards.tolerance(physics.angular_vel(), margin=5).min()
            small_velocity = (1 + small_velocity) / 2
            return upright.mean() * small_control * small_velocity * centered

    def get_reward(self, physics):
        """Returns a sparse or a smooth reward, as specified in the constructor."""
        return self._get_reward(physics, sparse=self._sparse)



# Copyright 2017 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

"""Humanoid Domain."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import collections

from dm_control import mujoco
from dm_control.rl import control
from . import base
from . import common
from dm_control.suite.utils import randomizers
from dm_control.utils import containers
from dm_control.utils import rewards
import numpy as np

_DEFAULT_TIME_LIMIT = 25
_CONTROL_TIMESTEP = 0.025

# Height of head above which stand reward is 1.
_STAND_HEIGHT = 1.4

# Horizontal speeds above which move reward is 1.
_WALK_SPEED = 1
_RUN_SPEED = 10


SUITE = containers.TaggedTasks()


def get_model_and_assets():
    """Returns a tuple containing the model XML string and a dict of assets."""
    return common.read_model("humanoid.xml"), common.ASSETS


@SUITE.add("benchmarking")
def stand(time_limit=_DEFAULT_TIME_LIMIT, random=None, environment_kwargs=None):
    """Returns the Stand task."""
    physics = Physics.from_xml_string(*get_model_and_assets())
    task = Humanoid(move_speed=0, pure_state=False, random=random)
    environment_kwargs = environment_kwargs or {}
    return control.Environment(
        physics,
        task,
        time_limit=time_limit,
        control_timestep=_CONTROL_TIMESTEP,
        **environment_kwargs
    )


@SUITE.add("benchmarking")
def walk(time_limit=_DEFAULT_TIME_LIMIT, random=None, environment_kwargs=None):
    """Returns the Walk task."""
    physics = Physics.from_xml_string(*get_model_and_assets())
    task = Humanoid(move_speed=_WALK_SPEED, pure_state=False, random=random)
    environment_kwargs = environment_kwargs or {}
    return control.Environment(
        physics,
        task,
        time_limit=time_limit,
        control_timestep=_CONTROL_TIMESTEP,
        **environment_kwargs
    )


@SUITE.add("benchmarking")
def run(time_limit=_DEFAULT_TIME_LIMIT, random=None, environment_kwargs=None):
    """Returns the Run task."""
    physics = Physics.from_xml_string(*get_model_and_assets())
    task = Humanoid(move_speed=_RUN_SPEED, pure_state=False, random=random)
    environment_kwargs = environment_kwargs or {}
    return control.Environment(
        physics,
        task,
        time_limit=time_limit,
        control_timestep=_CONTROL_TIMESTEP,
        **environment_kwargs
    )


@SUITE.add()
def run_pure_state(
    time_limit=_DEFAULT_TIME_LIMIT, random=None, environment_kwargs=None
):
    """Returns the Run task."""
    physics = Physics.from_xml_string(*get_model_and_assets())
    task = Humanoid(move_speed=_RUN_SPEED, pure_state=True, random=random)
    environment_kwargs = environment_kwargs or {}
    return control.Environment(
        physics,
        task,
        time_limit=time_limit,
        control_timestep=_CONTROL_TIMESTEP,
        **environment_kwargs
    )


class Physics(mujoco.Physics):
    """Physics simulation with additional features for the Walker domain."""

    def torso_upright(self):
        """Returns projection from z-axes of torso to the z-axes of world."""
        return self.named.data.xmat["torso", "zz"]

    def head_height(self):
        """Returns the height of the torso."""
        return self.named.data.xpos["head", "z"]

    def center_of_mass_position(self):
        """Returns position of the center-of-mass."""
        return self.named.data.subtree_com["torso"].copy()

    def center_of_mass_velocity(self):
        """Returns the velocity of the center-of-mass."""
        return self.named.data.sensordata["torso_subtreelinvel"].copy()

    def torso_vertical_orientation(self):
        """Returns the z-projection of the torso orientation matrix."""
        return self.named.data.xmat["torso", ["zx", "zy", "zz"]]

    def joint_angles(self):
        """Returns the state without global orientation or position."""
        return self.data.qpos[7:].copy()  # Skip the 7 DoFs of the free root joint.

    def extremities(self):
        """Returns end effector positions in egocentric frame."""
        torso_frame = self.named.data.xmat["torso"].reshape(3, 3)
        torso_pos = self.named.data.xpos["torso"]
        positions = []
        for side in ("left_", "right_"):
            for limb in ("hand", "foot"):
                torso_to_limb = self.named.data.xpos[side + limb] - torso_pos
                positions.append(torso_to_limb.dot(torso_frame))
        return np.hstack(positions)


class Humanoid(base.Task):
    """A humanoid task."""

    def __init__(self, move_speed, pure_state, random=None):
        """Initializes an instance of `Humanoid`.

        Args:
          move_speed: A float. If this value is zero, reward is given simply for
            standing up. Otherwise this specifies a target horizontal velocity for
            the walking task.
          pure_state: A bool. Whether the observations consist of the pure MuJoCo
            state or includes some useful features thereof.
          random: Optional, either a `numpy.random.RandomState` instance, an
            integer seed for creating a new `RandomState`, or None to select a seed
            automatically (default).
        """
        self._move_speed = move_speed
        self._pure_state = pure_state
        super(Humanoid, self).__init__(random=random)

    def initialize_episode(self, physics):
        """Sets the state of the environment at the start of each episode.

        Args:
          physics: An instance of `Physics`.

        """
        # Find a collision-free random initial configuration.
        penetrating = True
        while penetrating:
            randomizers.randomize_limited_and_rotational_joints(physics, self.random)
            # Check for collisions.
            physics.after_reset()
            penetrating = physics.data.ncon > 0
        super(Humanoid, self).initialize_episode(physics)

    def get_observation(self, physics):
        """Returns either the pure state or a set of egocentric features."""
        obs = collections.OrderedDict()
        if self._pure_state:
            obs["position"] = physics.position()
            obs["velocity"] = physics.velocity()
        else:
            obs["joint_angles"] = physics.joint_angles()
            obs["head_height"] = physics.head_height()
            obs["extremities"] = physics.extremities()
            obs["torso_vertical"] = physics.torso_vertical_orientation()
            obs["com_velocity"] = physics.center_of_mass_velocity()
            obs["velocity"] = physics.velocity()
        return obs

    def get_reward(self, physics):
        """Returns a reward to the agent."""
        standing = rewards.tolerance(
            physics.head_height(),
            bounds=(_STAND_HEIGHT, float("inf")),
            margin=_STAND_HEIGHT / 4,
        )
        upright = rewards.tolerance(
            physics.torso_upright(),
            bounds=(0.9, float("inf")),
            sigmoid="linear",
            margin=1.9,
            value_at_margin=0,
        )
        stand_reward = standing * upright
        small_control = rewards.tolerance(
            physics.control(), margin=1, value_at_margin=0, sigmoid="quadratic"
        ).mean()
        small_control = (4 + small_control) / 5
        if self._move_speed == 0:
            horizontal_velocity = physics.center_of_mass_velocity()[[0, 1]]
            dont_move = rewards.tolerance(horizontal_velocity, margin=2).mean()
            return small_control * stand_reward * dont_move
        else:
            com_velocity = np.linalg.norm(physics.center_of_mass_velocity()[[0, 1]])
            move = rewards.tolerance(
                com_velocity,
                bounds=(self._move_speed, float("inf")),
                margin=self._move_speed,
                value_at_margin=0,
                sigmoid="linear",
            )
            move = (5 * move + 1) / 6
            return small_control * stand_reward * move



# Copyright 2017 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

"""Procedurally generated Swimmer domain."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import collections

from dm_control import mujoco
from dm_control.rl import control
from . import base
from . import common
from dm_control.suite.utils import randomizers
from dm_control.utils import containers
from dm_control.utils import rewards
from lxml import etree
import numpy as np
from six.moves import range

_DEFAULT_TIME_LIMIT = 30
_CONTROL_TIMESTEP = 0.03  # (Seconds)

SUITE = containers.TaggedTasks()


def get_model_and_assets(n_joints):
    """Returns a tuple containing the model XML string and a dict of assets.

    Args:
      n_joints: An integer specifying the number of joints in the swimmer.

    Returns:
      A tuple `(model_xml_string, assets)`, where `assets` is a dict consisting of
      `{filename: contents_string}` pairs.
    """
    return _make_model(n_joints), common.ASSETS


@SUITE.add("benchmarking")
def swimmer6(time_limit=_DEFAULT_TIME_LIMIT, random=None, environment_kwargs=None):
    """Returns a 6-link swimmer."""
    return _make_swimmer(
        6, time_limit, random=random, environment_kwargs=environment_kwargs
    )


@SUITE.add("benchmarking")
def swimmer15(time_limit=_DEFAULT_TIME_LIMIT, random=None, environment_kwargs=None):
    """Returns a 15-link swimmer."""
    return _make_swimmer(
        15, time_limit, random=random, environment_kwargs=environment_kwargs
    )


def swimmer(
    n_links=3, time_limit=_DEFAULT_TIME_LIMIT, random=None, environment_kwargs=None
):
    """Returns a swimmer with n links."""
    return _make_swimmer(
        n_links, time_limit, random=random, environment_kwargs=environment_kwargs
    )


def _make_swimmer(
    n_joints, time_limit=_DEFAULT_TIME_LIMIT, random=None, environment_kwargs=None
):
    """Returns a swimmer control environment."""
    model_string, assets = get_model_and_assets(n_joints)
    physics = Physics.from_xml_string(model_string, assets=assets)
    task = Swimmer(random=random)
    environment_kwargs = environment_kwargs or {}
    return control.Environment(
        physics,
        task,
        time_limit=time_limit,
        control_timestep=_CONTROL_TIMESTEP,
        **environment_kwargs
    )


def _make_model(n_bodies):
    """Generates an xml string defining a swimmer with `n_bodies` bodies."""
    if n_bodies < 3:
        raise ValueError("At least 3 bodies required. Received {}".format(n_bodies))
    mjcf = etree.fromstring(common.read_model("swimmer.xml"))
    head_body = mjcf.find("./worldbody/body")
    actuator = etree.SubElement(mjcf, "actuator")
    sensor = etree.SubElement(mjcf, "sensor")

    parent = head_body
    for body_index in range(n_bodies - 1):
        site_name = "site_{}".format(body_index)
        child = _make_body(body_index=body_index)
        child.append(etree.Element("site", name=site_name))
        joint_name = "joint_{}".format(body_index)
        joint_limit = 360.0 / n_bodies
        joint_range = "{} {}".format(-joint_limit, joint_limit)
        child.append(etree.Element("joint", {"name": joint_name, "range": joint_range}))
        motor_name = "motor_{}".format(body_index)
        actuator.append(etree.Element("motor", name=motor_name, joint=joint_name))
        velocimeter_name = "velocimeter_{}".format(body_index)
        sensor.append(
            etree.Element("velocimeter", name=velocimeter_name, site=site_name)
        )
        gyro_name = "gyro_{}".format(body_index)
        sensor.append(etree.Element("gyro", name=gyro_name, site=site_name))
        parent.append(child)
        parent = child

    # Move tracking cameras further away from the swimmer according to its length.
    cameras = mjcf.findall("./worldbody/body/camera")
    scale = n_bodies / 6.0
    for cam in cameras:
        if cam.get("mode") == "trackcom":
            old_pos = cam.get("pos").split(" ")
            new_pos = " ".join([str(float(dim) * scale) for dim in old_pos])
            cam.set("pos", new_pos)

    return etree.tostring(mjcf, pretty_print=True)


def _make_body(body_index):
    """Generates an xml string defining a single physical body."""
    body_name = "segment_{}".format(body_index)
    visual_name = "visual_{}".format(body_index)
    inertial_name = "inertial_{}".format(body_index)
    body = etree.Element("body", name=body_name)
    body.set("pos", "0 .1 0")
    etree.SubElement(body, "geom", {"class": "visual", "name": visual_name})
    etree.SubElement(body, "geom", {"class": "inertial", "name": inertial_name})
    return body


class Physics(mujoco.Physics):
    """Physics simulation with additional features for the swimmer domain."""

    def nose_to_target(self):
        """Returns a vector from nose to target in local coordinate of the head."""
        nose_to_target = (
            self.named.data.geom_xpos["target"] - self.named.data.geom_xpos["nose"]
        )
        head_orientation = self.named.data.xmat["head"].reshape(3, 3)
        return nose_to_target.dot(head_orientation)[:2]

    def nose_to_target_dist(self):
        """Returns the distance from the nose to the target."""
        return np.linalg.norm(self.nose_to_target())

    def body_velocities(self):
        """Returns local body velocities: x,y linear, z rotational."""
        xvel_local = self.data.sensordata[12:].reshape((-1, 6))
        vx_vy_wz = [0, 1, 5]  # Indices for linear x,y vels and rotational z vel.
        return xvel_local[:, vx_vy_wz].ravel()

    def joints(self):
        """Returns all internal joint angles (excluding root joints)."""
        return self.data.qpos[3:].copy()


class Swimmer(base.Task):
    """A swimmer `Task` to reach the target or just swim."""

    def __init__(self, random=None):
        """Initializes an instance of `Swimmer`.

        Args:
          random: Optional, either a `numpy.random.RandomState` instance, an
            integer seed for creating a new `RandomState`, or None to select a seed
            automatically (default).
        """
        super(Swimmer, self).__init__(random=random)

    def initialize_episode(self, physics):
        """Sets the state of the environment at the start of each episode.

        Initializes the swimmer orientation to [-pi, pi) and the relative joint
        angle of each joint uniformly within its range.

        Args:
          physics: An instance of `Physics`.
        """
        # Random joint angles:
        randomizers.randomize_limited_and_rotational_joints(physics, self.random)
        # Random target position.
        close_target = self.random.rand() < 0.2  # Probability of a close target.
        target_box = 0.3 if close_target else 2
        xpos, ypos = self.random.uniform(-target_box, target_box, size=2)
        physics.named.model.geom_pos["target", "x"] = xpos
        physics.named.model.geom_pos["target", "y"] = ypos
        physics.named.model.light_pos["target_light", "x"] = xpos
        physics.named.model.light_pos["target_light", "y"] = ypos

        super(Swimmer, self).initialize_episode(physics)

    def get_observation(self, physics):
        """Returns an observation of joint angles, body velocities and target."""
        obs = collections.OrderedDict()
        obs["joints"] = physics.joints()
        obs["to_target"] = physics.nose_to_target()
        obs["body_velocities"] = physics.body_velocities()
        return obs

    def get_reward(self, physics):
        """Returns a smooth reward."""
        target_size = physics.named.model.geom_size["target", 0]
        return rewards.tolerance(
            physics.nose_to_target_dist(),
            bounds=(0, target_size),
            margin=5 * target_size,
            sigmoid="long_tail",
        )



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved
from mtenv.wrappers.ntasks import NTasks  # noqa: F401
from mtenv.wrappers.ntasks_id import NTasksId  # noqa: F401
from mtenv.wrappers.sample_random_task import SampleRandomTask  # noqa: F401



# Copyright 2017 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

"""Base class for tasks in the Control Suite."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from dm_control import mujoco
from dm_control.rl import control

import numpy as np


class Task(control.Task):
    """Base class for tasks in the Control Suite.

    Actions are mapped directly to the states of MuJoCo actuators: each element of
    the action array is used to set the control input for a single actuator. The
    ordering of the actuators is the same as in the corresponding MJCF XML file.

    Attributes:
      random: A `numpy.random.RandomState` instance. This should be used to
        generate all random variables associated with the task, such as random
        starting states, observation noise* etc.

    *If sensor noise is enabled in the MuJoCo model then this will be generated
    using MuJoCo's internal RNG, which has its own independent state.
    """

    def __init__(self, random=None):
        """Initializes a new continuous control task.

        Args:
          random: Optional, either a `numpy.random.RandomState` instance, an integer
            seed for creating a new `RandomState`, or None to select a seed
            automatically (default).
        """
        if not isinstance(random, np.random.RandomState):
            random = np.random.RandomState(random)
        self._random = random
        self._visualize_reward = False

    @property
    def random(self):
        """Task-specific `numpy.random.RandomState` instance."""
        return self._random

    def action_spec(self, physics):
        """Returns a `BoundedArraySpec` matching the `physics` actuators."""
        return mujoco.action_spec(physics)

    def initialize_episode(self, physics):
        """Resets geom colors to their defaults after starting a new episode.

        Subclasses of `base.Task` must delegate to this method after performing
        their own initialization.

        Args:
          physics: An instance of `mujoco.Physics`.
        """
        self.after_step(physics)

    def before_step(self, action, physics):
        """Sets the control signal for the actuators to values in `action`."""
        # Support legacy internal code.
        action = getattr(action, "continuous_actions", action)
        physics.set_control(action)

    def after_step(self, physics):
        """Modifies colors according to the reward."""
        if self._visualize_reward:
            reward = np.clip(self.get_reward(physics), 0.0, 1.0)
            _set_reward_colors(physics, reward)

    @property
    def visualize_reward(self):
        return self._visualize_reward

    @visualize_reward.setter
    def visualize_reward(self, value):
        if not isinstance(value, bool):
            raise ValueError("Expected a boolean, got {}.".format(type(value)))
        self._visualize_reward = value


_MATERIALS = ["self", "effector", "target"]
_DEFAULT = [name + "_default" for name in _MATERIALS]
_HIGHLIGHT = [name + "_highlight" for name in _MATERIALS]


def _set_reward_colors(physics, reward):
    """Sets the highlight, effector and target colors according to the reward."""
    assert 0.0 <= reward <= 1.0
    colors = physics.named.model.mat_rgba
    default = colors[_DEFAULT]
    highlight = colors[_HIGHLIGHT]
    blend_coef = reward ** 4  # Better color distinction near high rewards.
    colors[_MATERIALS] = blend_coef * highlight + (1.0 - blend_coef) * default



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved
"""Wrapper to fix the number of tasks in an existing multitask environment."""

from typing import List

from mtenv import MTEnv
from mtenv.utils.types import TaskStateType
from mtenv.wrappers.multitask import MultiTask


class NTasks(MultiTask):
    def __init__(self, env: MTEnv, n_tasks: int):
        """Wrapper to fix the number of tasks in an existing multitask
        environment to `n_tasks`.

        Each task is sampled in this fixed set of `n_tasks`.

        Args:
            env (MTEnv): Multitask environment to wrap over.
            n_tasks (int): Number of tasks to sample.
        """
        super().__init__(env=env)
        self.n_tasks = n_tasks
        self.tasks: List[TaskStateType]
        self._are_tasks_set = False

    def sample_task_state(self) -> TaskStateType:
        """Sample a `task_state` from the set of `n_tasks` tasks.

        `task_state` contains all the information that the environment
        needs to switch to any other task.

        The subclasses, extending this class, should ensure that the task
        seed is set (by calling `seed(int)`) before invoking this
        method (for reproducibility). It can be done by invoking
        `self.assert_task_seed_is_set()`.

        Returns:
            TaskStateType: For more information on `task_state`,
            refer :ref:`task_state`.
        """
        self.assert_task_seed_is_set()
        if not self._are_tasks_set:
            self.tasks = [self.env.sample_task_state() for _ in range(self.n_tasks)]
            self._are_tasks_set = True

        # The assert statement (at the start of the function) ensures that self.np_random_task
        # is not None. Mypy is raising the warning incorrectly.
        id_task = self.np_random_task.randint(self.n_tasks)  # type: ignore[union-attr]
        return self.tasks[id_task]

    def reset_task_state(self) -> None:
        """Sample a new task_state from the set of `n_tasks` tasks and
        set the environment to that `task_state`.

        For more information on `task_state`, refer :ref:`task_state`.
        """
        self.set_task_state(task_state=self.sample_task_state())



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved
from typing import Optional, Tuple

from gym.utils import seeding
from numpy.random import RandomState


def np_random(seed: Optional[int]) -> Tuple[RandomState, int]:
    """Set the seed for numpy's random generator.

    Args:
        seed (Optional[int]):

    Returns:
        Tuple[RandomState, int]: Returns a tuple of random state and seed.
    """
    rng, seed = seeding.np_random(seed)
    assert isinstance(seed, int)
    return rng, seed



# This code is taken from: https://raw.githubusercontent.com/rlworkgroup/garage/af57bf9c6b10cd733cb0fa9bfe3abd0ba239fd6e/src/garage/envs/normalized_env.py
#
# """"An environment wrapper that normalizes action, observation and reward."""
# type: ignore
import gym
import gym.spaces
import gym.spaces.utils
import numpy as np


class NormalizedEnvWrapper(gym.Wrapper):
    """An environment wrapper for normalization.

    This wrapper normalizes action, and optionally observation and reward.

    Args:
        env (garage.envs.GarageEnv): An environment instance.
        scale_reward (float): Scale of environment reward.
        normalize_obs (bool): If True, normalize observation.
        normalize_reward (bool): If True, normalize reward. scale_reward is
            applied after normalization.
        expected_action_scale (float): Assuming action falls in the range of
            [-expected_action_scale, expected_action_scale] when normalize it.
        flatten_obs (bool): Flatten observation if True.
        obs_alpha (float): Update rate of moving average when estimating the
            mean and variance of observations.
        reward_alpha (float): Update rate of moving average when estimating the
            mean and variance of rewards.

    """

    def __init__(
        self,
        env,
        scale_reward=1.0,
        normalize_obs=False,
        normalize_reward=False,
        expected_action_scale=1.0,
        flatten_obs=True,
        obs_alpha=0.001,
        reward_alpha=0.001,
    ):
        super().__init__(env)

        self._scale_reward = scale_reward
        self._normalize_obs = normalize_obs
        self._normalize_reward = normalize_reward
        self._expected_action_scale = expected_action_scale
        self._flatten_obs = flatten_obs

        self._obs_alpha = obs_alpha
        flat_obs_dim = gym.spaces.utils.flatdim(env.observation_space)
        self._obs_mean = np.zeros(flat_obs_dim)
        self._obs_var = np.ones(flat_obs_dim)

        self._reward_alpha = reward_alpha
        self._reward_mean = 0.0
        self._reward_var = 1.0

    def _update_obs_estimate(self, obs):
        flat_obs = gym.spaces.utils.flatten(self.env.observation_space, obs)
        self._obs_mean = (
            1 - self._obs_alpha
        ) * self._obs_mean + self._obs_alpha * flat_obs
        self._obs_var = (
            1 - self._obs_alpha
        ) * self._obs_var + self._obs_alpha * np.square(flat_obs - self._obs_mean)

    def _update_reward_estimate(self, reward):
        self._reward_mean = (
            1 - self._reward_alpha
        ) * self._reward_mean + self._reward_alpha * reward
        self._reward_var = (
            1 - self._reward_alpha
        ) * self._reward_var + self._reward_alpha * np.square(
            reward - self._reward_mean
        )

    def _apply_normalize_obs(self, obs):
        """Compute normalized observation.

        Args:
            obs (np.ndarray): Observation.

        Returns:
            np.ndarray: Normalized observation.

        """
        self._update_obs_estimate(obs)
        flat_obs = gym.spaces.utils.flatten(self.env.observation_space, obs)
        normalized_obs = (flat_obs - self._obs_mean) / (np.sqrt(self._obs_var) + 1e-8)
        if not self._flatten_obs:
            normalized_obs = gym.spaces.utils.unflatten(
                self.env.observation_space, normalized_obs
            )
        return normalized_obs

    def _apply_normalize_reward(self, reward):
        """Compute normalized reward.

        Args:
            reward (float): Reward.

        Returns:
            float: Normalized reward.

        """
        self._update_reward_estimate(reward)
        return reward / (np.sqrt(self._reward_var) + 1e-8)

    def reset(self, **kwargs):
        """Reset environment.

        Args:
            **kwargs: Additional parameters for reset.

        Returns:
            tuple:
                * observation (np.ndarray): The observation of the environment.
                * reward (float): The reward acquired at this time step.
                * done (boolean): Whether the environment was completed at this
                    time step.
                * infos (dict): Environment-dependent additional information.

        """
        ret = self.env.reset(**kwargs)
        if self._normalize_obs:
            return self._apply_normalize_obs(ret)
        else:
            return ret

    def step(self, action):
        """Feed environment with one step of action and get result.

        Args:
            action (np.ndarray): An action fed to the environment.

        Returns:
            tuple:
                * observation (np.ndarray): The observation of the environment.
                * reward (float): The reward acquired at this time step.
                * done (boolean): Whether the environment was completed at this
                    time step.
                * infos (dict): Environment-dependent additional information.

        """
        if isinstance(self.action_space, gym.spaces.Box):
            # rescale the action when the bounds are not inf
            lb, ub = self.action_space.low, self.action_space.high
            if np.all(lb != -np.inf) and np.all(ub != -np.inf):
                scaled_action = lb + (action + self._expected_action_scale) * (
                    0.5 * (ub - lb) / self._expected_action_scale
                )
                scaled_action = np.clip(scaled_action, lb, ub)
            else:
                scaled_action = action
        else:
            scaled_action = action
        try:
            next_obs, reward, done, info = self.env.step(scaled_action)
        except Exception as e:
            print(e)

        if self._normalize_obs:
            next_obs = self._apply_normalize_obs(next_obs)
        if self._normalize_reward:
            reward = self._apply_normalize_reward(reward)

        return next_obs, reward * self._scale_reward, done, info



# Copyright 2017 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

"""Cheetah Domain."""

from __future__ import absolute_import, division, print_function

import collections

from dm_control import mujoco
from dm_control.rl import control
from dm_control.utils import containers, rewards
from . import base, common

# How long the simulation will run, in seconds.
_DEFAULT_TIME_LIMIT = 10

# Running speed above which reward is 1.
_RUN_SPEED = 10

SUITE = containers.TaggedTasks()


def get_model_and_assets(xml_file_id):
    """Returns a tuple containing the model XML string and a dict of assets."""
    if xml_file_id is not None:
        filename = f"cheetah_{xml_file_id}.xml"
        print(filename)
    else:
        filename = f"cheetah.xml"
    return common.read_model(filename), common.ASSETS


@SUITE.add("benchmarking")
def run(
    time_limit=_DEFAULT_TIME_LIMIT,
    xml_file_id=None,
    random=None,
    environment_kwargs=None,
):
    """Returns the run task."""
    physics = Physics.from_xml_string(*get_model_and_assets(xml_file_id))
    task = Cheetah(random=random)
    environment_kwargs = environment_kwargs or {}
    return control.Environment(
        physics, task, time_limit=time_limit, **environment_kwargs
    )


class Physics(mujoco.Physics):
    """Physics simulation with additional features for the Cheetah domain."""

    def speed(self):
        """Returns the horizontal speed of the Cheetah."""
        return self.named.data.sensordata["torso_subtreelinvel"][0]


class Cheetah(base.Task):
    """A `Task` to train a running Cheetah."""

    def initialize_episode(self, physics):
        """Sets the state of the environment at the start of each episode."""
        # The indexing below assumes that all joints have a single DOF.
        assert physics.model.nq == physics.model.njnt
        is_limited = physics.model.jnt_limited == 1
        lower, upper = physics.model.jnt_range[is_limited].T
        physics.data.qpos[is_limited] = self.random.uniform(lower, upper)

        # Stabilize the model before the actual simulation.
        for _ in range(200):
            physics.step()

        physics.data.time = 0
        self._timeout_progress = 0
        super(Cheetah, self).initialize_episode(physics)

    def get_observation(self, physics):
        """Returns an observation of the state, ignoring horizontal position."""
        obs = collections.OrderedDict()
        # Ignores horizontal position to maintain translational invariance.
        obs["position"] = physics.data.qpos[1:].copy()
        obs["velocity"] = physics.velocity()
        return obs

    def get_reward(self, physics):
        """Returns a reward to the agent."""
        return rewards.tolerance(
            physics.speed(),
            bounds=(_RUN_SPEED, float("inf")),
            margin=_RUN_SPEED,
            value_at_margin=0,
            sigmoid="linear",
        )



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved

from copy import deepcopy

from mtenv.envs.registration import register

# Control Task
# ----------------------------------------

register(
    id="MT-CartPole-v0",
    entry_point="mtenv.envs.control.cartpole:MTCartPole",
    test_kwargs={
        # "valid_env_kwargs": [],
        "invalid_env_kwargs": [],
    },
)


register(
    id="MT-TabularMDP-v0",
    entry_point="mtenv.envs.tabular_mdp.tmdp:UniformTMDP",
    kwargs={"n_states": 4, "n_actions": 5},
    test_kwargs={
        "valid_env_kwargs": [{"n_states": 3, "n_actions": 2}],
        "invalid_env_kwargs": [],
    },
)

register(
    id="MT-Acrobat-v0",
    entry_point="mtenv.envs.control.acrobot:MTAcrobot",
    test_kwargs={
        # "valid_env_kwargs": [],
        "invalid_env_kwargs": [],
    },
)

register(
    id="MT-TwoGoalMaze-v0",
    entry_point="mtenv.envs.mpte.two_goal_maze_env:build_two_goal_maze_env",
    kwargs={"size_x": 3, "size_y": 3, "task_seed": 169, "n_tasks": 100},
    test_kwargs={
        # "valid_env_kwargs": [],
        "invalid_env_kwargs": [],
    },
)


# remove it before making the repo public.
default_kwargs = {
    "seed": 1,
    "visualize_reward": False,
    "from_pixels": True,
    "height": 84,
    "width": 84,
    "frame_skip": 2,
    "frame_stack": 3,
    "sticky_observation_cfg": {},
    "initial_task_state": 1,
}

for domain_name, task_name, prefix in [
    ("finger", "spin", "size"),
    ("cheetah", "run", "torso_length"),
    ("walker", "walk", "friction"),
    ("walker", "walk", "len"),
]:
    file_ids = list(range(1, 11))
    kwargs = deepcopy(default_kwargs)
    kwargs["domain_name"] = domain_name
    kwargs["task_name"] = task_name
    kwargs["xml_file_ids"] = [f"{prefix}_{i}" for i in file_ids]
    register(
        id=f"MT-HiPBMDP-{domain_name.capitalize()}-{task_name.capitalize()}-vary-{prefix.replace('_', '-')}-v0",
        entry_point="mtenv.envs.hipbmdp.env:build",
        kwargs=kwargs,
        test_kwargs={
            # "valid_env_kwargs": [],
            # "invalid_env_kwargs": [],
        },
    )


default_kwargs = {
    "benchmark": None,
    "benchmark_name": "MT10",
    "env_id_to_task_map": None,
    "should_perform_reward_normalization": True,
    "num_copies_per_env": 1,
    "initial_task_state": 1,
}

for benchmark_name in [("MT10"), ("MT50")]:
    kwargs = deepcopy(default_kwargs)
    kwargs["benchmark_name"] = benchmark_name
    register(
        id=f"MT-MetaWorld-{benchmark_name}-v0",
        entry_point="mtenv.envs.metaworld.env:build",
        kwargs=kwargs,
        test_kwargs={
            # "valid_env_kwargs": [],
            # "invalid_env_kwargs": [],
        },
    )

kwargs = {
    "benchmark": None,
    "benchmark_name": "MT1",
    "env_id_to_task_map": None,
    "should_perform_reward_normalization": True,
    "task_name": "pick-place-v1",
    "num_copies_per_env": 1,
    "initial_task_state": 0,
}
register(
    id=f'MT-MetaWorld-{kwargs["benchmark_name"]}-v0',
    entry_point="mtenv.envs.metaworld.env:build",
    kwargs=kwargs,
    test_kwargs={
        # "valid_env_kwargs": [],
        # "invalid_env_kwargs": [],
    },
)



# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the  LICENSE file in the root directory of this source tree.


import numpy as np
from gym import spaces
from numpy import cos, pi, sin

from mtenv import MTEnv
from mtenv.utils import seeding

__copyright__ = "Copyright 2013, RLPy http://acl.mit.edu/RLPy"
__credits__ = [
    "Alborz Geramifard",
    "Robert H. Klein",
    "Christoph Dann",
    "William Dabney",
    "Jonathan P. How",
]
__license__ = "BSD 3-Clause"
__author__ = "Christoph Dann <cdann@cdann.de>"

# SOURCE:
# https://github.com/rlpy/rlpy/blob/master/rlpy/Domains/Acrobot.py


class MTAcrobot(MTEnv):
    """A acrobot environment with varying characteristics
    The task descriptor is composed of values between -1 and +1 and mapped to acrobot physical characcteristics in the
    self._mu_to_vars function.


    """

    metadata = {"render.modes": ["human", "rgb_array"], "video.frames_per_second": 15}

    dt = 0.2

    def _mu_to_vars(self, mu):
        self.LINK_LENGTH_1 = 1.0 + mu[0] * 0.5
        self.LINK_LENGTH_2 = 1.0 + mu[1] * 0.5
        self.LINK_MASS_1 = 1.0 + mu[2] * 0.5
        self.LINK_MASS_2 = 1.0 + mu[3] * 0.5
        self.LINK_COM_POS_1 = 0.5
        self.LINK_COM_POS_2 = 0.5
        if mu[6] > 0:
            self.AVAIL_TORQUE = [-1.0, 0.0, 1.0]
        else:
            self.AVAIL_TORQUE = [1.0, 0.0, -1.0]
        self.LINK_MOI = 1.0

    torque_noise_max = 0.0
    MAX_VEL_1 = 4 * pi + pi
    MAX_VEL_2 = 9 * pi + 2 * pi

    #: use dynamics equations from the nips paper or the book
    book_or_nips = "book"
    action_arrow = None
    domain_fig = None
    actions_num = 3

    def __init__(self):
        self.viewer = None
        self.action_space = spaces.Discrete(3)
        self.state = None
        high = np.array(
            [1.5, 1.5, 1.5, 1.5, self.MAX_VEL_1, self.MAX_VEL_2], dtype=np.float32
        )
        low = -high
        observation_space = spaces.Box(low=low, high=high, dtype=np.float32)
        action_space = spaces.Discrete(3)
        high = np.array([1.0 for k in range(5)])
        task_space = spaces.Box(-high, high, dtype=np.float32)
        super().__init__(
            action_space=action_space,
            env_observation_space=observation_space,
            task_observation_space=task_space,
        )

    def step(self, a):
        self.t += 1
        self._mu_to_vars(self.task_state)
        s = self.state
        torque = self.AVAIL_TORQUE[a]

        # Add noise to the force action
        if self.torque_noise_max > 0:
            torque += self.np_random_env.uniform(
                -self.torque_noise_max, self.torque_noise_max
            )

        # Now, augment the state with our force action so it can be passed to
        # _dsdt
        s_augmented = np.append(s, torque)

        ns = rk4(self._dsdt, s_augmented, [0, self.dt])
        # only care about final timestep of integration returned by integrator
        ns = ns[-1]
        ns = ns[:4]  # omit action
        # ODEINT IS TOO SLOW!
        # ns_continuous = integrate.odeint(self._dsdt, self.s_continuous, [0, self.dt])
        # self.s_continuous = ns_continuous[-1] # We only care about the state
        # at the ''final timestep'', self.dt

        ns[0] = wrap(ns[0], -pi, pi)
        ns[1] = wrap(ns[1], -pi, pi)
        ns[2] = bound(ns[2], -self.MAX_VEL_1, self.MAX_VEL_1)
        ns[3] = bound(ns[3], -self.MAX_VEL_2, self.MAX_VEL_2)
        self.state = ns
        terminal = self._terminal()
        reward = -1.0 if not terminal else 0.0
        return (
            {"env_obs": self._get_obs(), "task_obs": self.get_task_obs()},
            reward,
            terminal,
            {},
        )

    def reset(self):
        self._mu_to_vars(self.task_state)
        self.state = self.np_random_env.uniform(low=-0.1, high=0.1, size=(4,))
        self.t = 0
        return {"env_obs": self._get_obs(), "task_obs": self.get_task_obs()}

    def get_task_obs(self):
        return self.task_state

    def get_task_state(self):
        return self.task_state

    def set_task_state(self, task_state):
        self.task_state = task_state

    def _get_obs(self):
        s = self.state
        return [cos(s[0]), sin(s[0]), cos(s[1]), sin(s[1]), s[2], s[3]]

    def _terminal(self):
        s = self.state
        return bool(-cos(s[0]) - cos(s[1] + s[0]) > 1.0)

    def _dsdt(self, s_augmented, t):
        m1 = self.LINK_MASS_1
        m2 = self.LINK_MASS_2
        l1 = self.LINK_LENGTH_1
        lc1 = self.LINK_COM_POS_1
        lc2 = self.LINK_COM_POS_2
        I1 = self.LINK_MOI
        I2 = self.LINK_MOI
        g = 9.8
        a = s_augmented[-1]
        s = s_augmented[:-1]
        theta1 = s[0]
        theta2 = s[1]
        dtheta1 = s[2]
        dtheta2 = s[3]
        d1 = (
            m1 * lc1 ** 2
            + m2 * (l1 ** 2 + lc2 ** 2 + 2 * l1 * lc2 * cos(theta2))
            + I1
            + I2
        )
        d2 = m2 * (lc2 ** 2 + l1 * lc2 * cos(theta2)) + I2
        phi2 = m2 * lc2 * g * cos(theta1 + theta2 - pi / 2.0)
        phi1 = (
            -m2 * l1 * lc2 * dtheta2 ** 2 * sin(theta2)
            - 2 * m2 * l1 * lc2 * dtheta2 * dtheta1 * sin(theta2)
            + (m1 * lc1 + m2 * l1) * g * cos(theta1 - pi / 2)
            + phi2
        )
        if self.book_or_nips == "nips":
            # the following line is consistent with the description in the
            # paper
            ddtheta2 = (a + d2 / d1 * phi1 - phi2) / (m2 * lc2 ** 2 + I2 - d2 ** 2 / d1)
        else:
            # the following line is consistent with the java implementation and the
            # book
            ddtheta2 = (
                a + d2 / d1 * phi1 - m2 * l1 * lc2 * dtheta1 ** 2 * sin(theta2) - phi2
            ) / (m2 * lc2 ** 2 + I2 - d2 ** 2 / d1)
        ddtheta1 = -(d2 * ddtheta2 + phi1) / d1
        return (dtheta1, dtheta2, ddtheta1, ddtheta2, 0.0)

    def seed(self, env_seed):
        self.np_random_env, seed = seeding.np_random(env_seed)
        return [seed]

    def seed_task(self, task_seed):
        self.np_random_task, seed = seeding.np_random(task_seed)
        return [seed]

    def sample_task_state(self):
        self.assert_task_seed_is_set()
        super().sample_task_state()
        new_task_state = [
            self.np_random_task.uniform(-1, 1),
            self.np_random_task.uniform(-1, 1),
            self.np_random_task.uniform(-1, 1),
            self.np_random_task.uniform(-1, 1),
            self.np_random_task.uniform(-1, 1),
            self.np_random_task.uniform(-1, 1),
            self.np_random_task.uniform(-1, 1),
        ]
        return new_task_state


def wrap(x, m, M):
    """
    :param x: a scalar
    :param m: minimum possible value in range
    :param M: maximum possible value in range
    Wraps ``x`` so m <= x <= M; but unlike ``bound()`` which
    truncates, ``wrap()`` wraps x around the coordinate system defined by m,M.\n
    For example, m = -180, M = 180 (degrees), x = 360 --> returns 0.
    """
    diff = M - m
    while x > M:
        x = x - diff
    while x < m:
        x = x + diff
    return x


def bound(x, m, M=None):
    """
    :param x: scalar
    Either have m as scalar, so bound(x,m,M) which returns m <= x <= M *OR*
    have m as length 2 vector, bound(x,m, <IGNORED>) returns m[0] <= x <= m[1].
    """
    if M is None:
        M = m[1]
        m = m[0]
    # bound x between min (m) and Max (M)
    return min(max(x, m), M)


def rk4(derivs, y0, t, *args, **kwargs):
    """
    Integrate 1D or ND system of ODEs using 4-th order Runge-Kutta.
    This is a toy implementation which may be useful if you find
    yourself stranded on a system w/o scipy.  Otherwise use
    :func:`scipy.integrate`.
    *y0*
        initial state vector
    *t*
        sample times
    *derivs*
        returns the derivative of the system and has the
        signature ``dy = derivs(yi, ti)``
    *args*
        additional arguments passed to the derivative function
    *kwargs*
        additional keyword arguments passed to the derivative function
    Example 1 ::
        ## 2D system
        def derivs6(x,t):
            d1 =  x[0] + 2*x[1]
            d2 =  -3*x[0] + 4*x[1]
            return (d1, d2)
        dt = 0.0005
        t = arange(0.0, 2.0, dt)
        y0 = (1,2)
        yout = rk4(derivs6, y0, t)
    Example 2::
        ## 1D system
        alpha = 2
        def derivs(x,t):
            return -alpha*x + exp(-t)
        y0 = 1
        yout = rk4(derivs, y0, t)
    If you have access to scipy, you should probably be using the
    scipy.integrate tools rather than this function.
    """

    try:
        Ny = len(y0)
    except TypeError:
        yout = np.zeros((len(t),), np.float_)
    else:
        yout = np.zeros((len(t), Ny), np.float_)

    yout[0] = y0

    for i in np.arange(len(t) - 1):

        thist = t[i]
        dt = t[i + 1] - thist
        dt2 = dt / 2.0
        y0 = yout[i]

        k1 = np.asarray(derivs(y0, thist, *args, **kwargs))
        k2 = np.asarray(derivs(y0 + dt2 * k1, thist + dt2, *args, **kwargs))
        k3 = np.asarray(derivs(y0 + dt2 * k2, thist + dt2, *args, **kwargs))
        k4 = np.asarray(derivs(y0 + dt * k3, thist + dt, *args, **kwargs))
        yout[i + 1] = y0 + dt / 6.0 * (k1 + 2 * k2 + 2 * k3 + k4)
    return yout


class Acrobot(MTAcrobot):
    """The original acrobot environment in the MTEnv fashion"""

    def __init__(self):
        super().__init__()

    def sample_task_state(self):
        self.assert_task_seed_is_set()
        super().sample_task_state()
        new_task_state = [
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
        ]
        return new_task_state


if __name__ == "__main__":
    env = MTAcrobot()
    env.seed(5)
    env.seed_task(15)
    env.reset_task_state()
    obs = env.reset()
    print(obs)
    done = False
    while not done:
        obs, rew, done, _ = env.step(np.random.randint(env.action_space.n))
        print(obs)



# Copyright 2017 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

"""Tests specific to the LQR domain."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import math
import unittest

# Internal dependencies.
from absl import logging

from absl.testing import absltest
from absl.testing import parameterized

from . import lqr
from . import lqr_solver

import numpy as np
from six.moves import range


class LqrTest(parameterized.TestCase):
    @parameterized.named_parameters(("lqr_2_1", lqr.lqr_2_1), ("lqr_6_2", lqr.lqr_6_2))
    def test_lqr_optimal_policy(self, make_env):
        env = make_env()
        p, k, beta = lqr_solver.solve(env)
        self.assertPolicyisOptimal(env, p, k, beta)

    @parameterized.named_parameters(("lqr_2_1", lqr.lqr_2_1), ("lqr_6_2", lqr.lqr_6_2))
    @unittest.skipUnless(
        condition=lqr_solver.sp,
        reason="scipy is not available, so non-scipy DARE solver is the default.",
    )
    def test_lqr_optimal_policy_no_scipy(self, make_env):
        env = make_env()
        old_sp = lqr_solver.sp
        try:
            lqr_solver.sp = None  # Force the solver to use the non-scipy code path.
            p, k, beta = lqr_solver.solve(env)
        finally:
            lqr_solver.sp = old_sp
        self.assertPolicyisOptimal(env, p, k, beta)

    def assertPolicyisOptimal(self, env, p, k, beta):
        tolerance = 1e-3
        n_steps = int(math.ceil(math.log10(tolerance) / math.log10(beta)))
        logging.info("%d timesteps for %g convergence.", n_steps, tolerance)
        total_loss = 0.0

        timestep = env.reset()
        initial_state = np.hstack(
            (timestep.observation["position"], timestep.observation["velocity"])
        )
        logging.info("Measuring total cost over %d steps.", n_steps)
        for _ in range(n_steps):
            x = np.hstack(
                (timestep.observation["position"], timestep.observation["velocity"])
            )
            # u = k*x is the optimal policy
            u = k.dot(x)
            total_loss += 1 - (timestep.reward or 0.0)
            timestep = env.step(u)

        logging.info("Analytical expected total cost is .5*x^T*p*x.")
        expected_loss = 0.5 * initial_state.T.dot(p).dot(initial_state)
        logging.info("Comparing measured and predicted costs.")
        np.testing.assert_allclose(expected_loss, total_loss, rtol=tolerance)


if __name__ == "__main__":
    absltest.main()



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved
# type: ignore
import codecs
import os.path
import subprocess
from pathlib import Path

import setuptools


def read(rel_path):
    here = os.path.abspath(os.path.dirname(__file__))
    with codecs.open(os.path.join(here, rel_path), "r") as fp:
        return fp.read()


def get_version(rel_path):
    for line in read(rel_path).splitlines():
        if line.startswith("__version__"):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    raise RuntimeError("Unable to find version string.")


def parse_dependency(filepath):
    dep_list = []
    for dep in open(filepath).read().splitlines():
        if dep.startswith("#"):
            continue
        key = "#egg="
        if key in dep:
            git_link, egg_name = dep.split(key)
            dep = f"{egg_name} @ {git_link}"
        dep_list.append(dep)
    return dep_list


base_requirements = parse_dependency("requirements/base.txt")
dev_requirements = base_requirements + parse_dependency("requirements/dev.txt")


extras_require = {}

for setup_path in Path("mtenv/envs").glob("**/setup.py"):
    env_path = setup_path.parent
    env_name = (
        subprocess.run(["python", setup_path, "--name"], stdout=subprocess.PIPE)
        .stdout.decode()
        .strip()
    )
    extras_require[env_name] = base_requirements + parse_dependency(
        f"{str(env_path)}/requirements.txt"
    )

extras_require["all"] = list(
    set([dep for requirements in extras_require.values() for dep in requirements])
)
extras_require["dev"] = dev_requirements

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="mtenv",
    version=get_version("mtenv/__init__.py"),
    author="Shagun Sodhani, Ludovic Denoyer, Pierre-Alexandre Kamienny, Olivier Delalleau",
    author_email="sshagunsodhani@gmail.com, denoyer@fb.com, pakamienny@fb.com, odelalleau@fb.com",
    description="MTEnv: MultiTask Environments for Reinforcement Learning",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    install_requires=base_requirements,
    url="https://github.com/facbookresearch/mtenv",
    packages=setuptools.find_packages(
        exclude=["*.tests", "*.tests.*", "tests.*", "tests", "docs", "docsrc"]
    ),
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
)



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved

from typing import Any, Dict, Optional

import dmc2gym
import numpy as np
from dmc2gym.wrappers import DMCWrapper as BaseDMCWrapper
from gym import spaces

import local_dm_control_suite as local_dmc_suite


class DMCWrapper(BaseDMCWrapper):
    def __init__(
        self,
        domain_name: str,
        task_name: str,
        task_kwargs: Any = None,
        visualize_reward: Optional[Dict[str, Any]] = None,
        from_pixels: bool = False,
        height=84,
        width: int = 84,
        camera_id: int = 0,
        frame_skip: int = 1,
        environment_kwargs: Any = None,
        channels_first: bool = True,
    ):
        """This wrapper is based on implementation from
        https://github.com/denisyarats/dmc2gym/blob/master/dmc2gym/wrappers.py#L37

        We extend the wrapper so that we can use the modified version of
        `dm_control_suite`.
        """
        assert (
            "random" in task_kwargs  # type: ignore [operator]
        ), "please specify a seed, for deterministic behaviour"
        self._from_pixels = from_pixels
        self._height = height
        self._width = width
        self._camera_id = camera_id
        self._frame_skip = frame_skip
        self._channels_first = channels_first
        if visualize_reward is None:
            visualize_reward = {}
        # create task
        self._env = local_dmc_suite.load(
            domain_name=domain_name,
            task_name=task_name,
            task_kwargs=task_kwargs,
            visualize_reward=visualize_reward,
            environment_kwargs=environment_kwargs,
        )

        # true and normalized action spaces
        self._true_action_space = dmc2gym.wrappers._spec_to_box(
            [self._env.action_spec()]
        )
        self._norm_action_space = spaces.Box(
            low=-1.0, high=1.0, shape=self._true_action_space.shape, dtype=np.float32
        )

        # create observation space
        if from_pixels:
            shape = [3, height, width] if channels_first else [height, width, 3]
            self._observation_space = spaces.Box(
                low=0, high=255, shape=shape, dtype=np.uint8
            )
        else:
            self._observation_space = dmc2gym.wrappers._spec_to_box(
                self._env.observation_spec().values()
            )

        self._state_space = dmc2gym.wrappers._spec_to_box(
            self._env.observation_spec().values()
        )

        self.current_state = None

        # set seed
        self.seed(seed=task_kwargs["random"])  # type: ignore [index]



# Copyright 2017 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

"""Point-mass domain."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import collections

from dm_control import mujoco
from dm_control.rl import control
from . import base
from . import common
from dm_control.suite.utils import randomizers
from dm_control.utils import containers
from dm_control.utils import rewards
import numpy as np

_DEFAULT_TIME_LIMIT = 20
SUITE = containers.TaggedTasks()


def get_model_and_assets():
    """Returns a tuple containing the model XML string and a dict of assets."""
    return common.read_model("point_mass.xml"), common.ASSETS


@SUITE.add("benchmarking", "easy")
def easy(time_limit=_DEFAULT_TIME_LIMIT, random=None, environment_kwargs=None):
    """Returns the easy point_mass task."""
    physics = Physics.from_xml_string(*get_model_and_assets())
    task = PointMass(randomize_gains=False, random=random)
    environment_kwargs = environment_kwargs or {}
    return control.Environment(
        physics, task, time_limit=time_limit, **environment_kwargs
    )


@SUITE.add()
def hard(time_limit=_DEFAULT_TIME_LIMIT, random=None, environment_kwargs=None):
    """Returns the hard point_mass task."""
    physics = Physics.from_xml_string(*get_model_and_assets())
    task = PointMass(randomize_gains=True, random=random)
    environment_kwargs = environment_kwargs or {}
    return control.Environment(
        physics, task, time_limit=time_limit, **environment_kwargs
    )


class Physics(mujoco.Physics):
    """physics for the point_mass domain."""

    def mass_to_target(self):
        """Returns the vector from mass to target in global coordinate."""
        return (
            self.named.data.geom_xpos["target"] - self.named.data.geom_xpos["pointmass"]
        )

    def mass_to_target_dist(self):
        """Returns the distance from mass to the target."""
        return np.linalg.norm(self.mass_to_target())


class PointMass(base.Task):
    """A point_mass `Task` to reach target with smooth reward."""

    def __init__(self, randomize_gains, random=None):
        """Initialize an instance of `PointMass`.

        Args:
          randomize_gains: A `bool`, whether to randomize the actuator gains.
          random: Optional, either a `numpy.random.RandomState` instance, an
            integer seed for creating a new `RandomState`, or None to select a seed
            automatically (default).
        """
        self._randomize_gains = randomize_gains
        super(PointMass, self).__init__(random=random)

    def initialize_episode(self, physics):
        """Sets the state of the environment at the start of each episode.

           If _randomize_gains is True, the relationship between the controls and
           the joints is randomized, so that each control actuates a random linear
           combination of joints.

        Args:
          physics: An instance of `mujoco.Physics`.
        """
        randomizers.randomize_limited_and_rotational_joints(physics, self.random)
        if self._randomize_gains:
            dir1 = self.random.randn(2)
            dir1 /= np.linalg.norm(dir1)
            # Find another actuation direction that is not 'too parallel' to dir1.
            parallel = True
            while parallel:
                dir2 = self.random.randn(2)
                dir2 /= np.linalg.norm(dir2)
                parallel = abs(np.dot(dir1, dir2)) > 0.9
            physics.model.wrap_prm[[0, 1]] = dir1
            physics.model.wrap_prm[[2, 3]] = dir2
        super(PointMass, self).initialize_episode(physics)

    def get_observation(self, physics):
        """Returns an observation of the state."""
        obs = collections.OrderedDict()
        obs["position"] = physics.position()
        obs["velocity"] = physics.velocity()
        return obs

    def get_reward(self, physics):
        """Returns a reward to the agent."""
        target_size = physics.named.model.geom_size["target", 0]
        near_target = rewards.tolerance(
            physics.mass_to_target_dist(), bounds=(0, target_size), margin=target_size
        )
        control_reward = rewards.tolerance(
            physics.control(), margin=1, value_at_margin=0, sigmoid="quadratic"
        ).mean()
        small_control = (control_reward + 4) / 5
        return near_target * small_control



# Copyright 2017 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

"""Finger Domain."""

from __future__ import absolute_import, division, print_function

import collections

import numpy as np
from dm_control import mujoco
from dm_control.rl import control
from dm_control.suite.utils import randomizers
from dm_control.utils import containers
from six.moves import range

from . import base, common

_DEFAULT_TIME_LIMIT = 20  # (seconds)
_CONTROL_TIMESTEP = 0.02  # (seconds)
# For TURN tasks, the 'tip' geom needs to enter a spherical target of sizes:
_EASY_TARGET_SIZE = 0.07
_HARD_TARGET_SIZE = 0.03
# Initial spin velocity for the Stop task.
_INITIAL_SPIN_VELOCITY = 100
# Spinning slower than this value (radian/second) is considered stopped.
_STOP_VELOCITY = 1e-6
# Spinning faster than this value (radian/second) is considered spinning.
_SPIN_VELOCITY = 15.0


SUITE = containers.TaggedTasks()


def get_model_and_assets(xml_file_id):
    """Returns a tuple containing the model XML string and a dict of assets."""
    if xml_file_id is not None:
        filename = f"finger_{xml_file_id}.xml"
    else:
        filename = f"finger.xml"
    return common.read_model(filename), common.ASSETS


@SUITE.add("benchmarking")
def spin(
    time_limit=_DEFAULT_TIME_LIMIT,
    xml_file_id=None,
    random=None,
    environment_kwargs=None,
):
    """Returns the Spin task."""
    physics = Physics.from_xml_string(*get_model_and_assets(xml_file_id))
    task = Spin(random=random)
    environment_kwargs = environment_kwargs or {}
    return control.Environment(
        physics,
        task,
        time_limit=time_limit,
        control_timestep=_CONTROL_TIMESTEP,
        **environment_kwargs,
    )


@SUITE.add("benchmarking")
def turn_easy(time_limit=_DEFAULT_TIME_LIMIT, random=None, environment_kwargs=None):
    """Returns the easy Turn task."""
    physics = Physics.from_xml_string(*get_model_and_assets())
    task = Turn(target_radius=_EASY_TARGET_SIZE, random=random)
    environment_kwargs = environment_kwargs or {}
    return control.Environment(
        physics,
        task,
        time_limit=time_limit,
        control_timestep=_CONTROL_TIMESTEP,
        **environment_kwargs,
    )


@SUITE.add("benchmarking")
def turn_hard(time_limit=_DEFAULT_TIME_LIMIT, random=None, environment_kwargs=None):
    """Returns the hard Turn task."""
    physics = Physics.from_xml_string(*get_model_and_assets())
    task = Turn(target_radius=_HARD_TARGET_SIZE, random=random)
    environment_kwargs = environment_kwargs or {}
    return control.Environment(
        physics,
        task,
        time_limit=time_limit,
        control_timestep=_CONTROL_TIMESTEP,
        **environment_kwargs,
    )


class Physics(mujoco.Physics):
    """Physics simulation with additional features for the Finger domain."""

    def touch(self):
        """Returns logarithmically scaled signals from the two touch sensors."""
        return np.log1p(self.named.data.sensordata[["touchtop", "touchbottom"]])

    def hinge_velocity(self):
        """Returns the velocity of the hinge joint."""
        return self.named.data.sensordata["hinge_velocity"]

    def tip_position(self):
        """Returns the (x,z) position of the tip relative to the hinge."""
        return (
            self.named.data.sensordata["tip"][[0, 2]]
            - self.named.data.sensordata["spinner"][[0, 2]]
        )

    def bounded_position(self):
        """Returns the positions, with the hinge angle replaced by tip position."""
        return np.hstack(
            (self.named.data.sensordata[["proximal", "distal"]], self.tip_position())
        )

    def velocity(self):
        """Returns the velocities (extracted from sensordata)."""
        return self.named.data.sensordata[
            ["proximal_velocity", "distal_velocity", "hinge_velocity"]
        ]

    def target_position(self):
        """Returns the (x,z) position of the target relative to the hinge."""
        return (
            self.named.data.sensordata["target"][[0, 2]]
            - self.named.data.sensordata["spinner"][[0, 2]]
        )

    def to_target(self):
        """Returns the vector from the tip to the target."""
        return self.target_position() - self.tip_position()

    def dist_to_target(self):
        """Returns the signed distance to the target surface, negative is inside."""
        return (
            np.linalg.norm(self.to_target()) - self.named.model.site_size["target", 0]
        )


class Spin(base.Task):
    """A Finger `Task` to spin the stopped body."""

    def __init__(self, random=None):
        """Initializes a new `Spin` instance.

        Args:
          random: Optional, either a `numpy.random.RandomState` instance, an
            integer seed for creating a new `RandomState`, or None to select a seed
            automatically (default).
        """
        super(Spin, self).__init__(random=random)

    def initialize_episode(self, physics):
        physics.named.model.site_rgba["target", 3] = 0
        physics.named.model.site_rgba["tip", 3] = 0
        physics.named.model.dof_damping["hinge"] = 0.03
        _set_random_joint_angles(physics, self.random)
        super(Spin, self).initialize_episode(physics)

    def get_observation(self, physics):
        """Returns state and touch sensors, and target info."""
        obs = collections.OrderedDict()
        obs["position"] = physics.bounded_position()
        obs["velocity"] = physics.velocity()
        obs["touch"] = physics.touch()
        return obs

    def get_reward(self, physics):
        """Returns a sparse reward."""
        return float(physics.hinge_velocity() <= -_SPIN_VELOCITY)


class Turn(base.Task):
    """A Finger `Task` to turn the body to a target angle."""

    def __init__(self, target_radius, random=None):
        """Initializes a new `Turn` instance.

        Args:
          target_radius: Radius of the target site, which specifies the goal angle.
          random: Optional, either a `numpy.random.RandomState` instance, an
            integer seed for creating a new `RandomState`, or None to select a seed
            automatically (default).
        """
        self._target_radius = target_radius
        super(Turn, self).__init__(random=random)

    def initialize_episode(self, physics):
        target_angle = self.random.uniform(-np.pi, np.pi)
        hinge_x, hinge_z = physics.named.data.xanchor["hinge", ["x", "z"]]
        radius = physics.named.model.geom_size["cap1"].sum()
        target_x = hinge_x + radius * np.sin(target_angle)
        target_z = hinge_z + radius * np.cos(target_angle)
        physics.named.model.site_pos["target", ["x", "z"]] = target_x, target_z
        physics.named.model.site_size["target", 0] = self._target_radius

        _set_random_joint_angles(physics, self.random)

        super(Turn, self).initialize_episode(physics)

    def get_observation(self, physics):
        """Returns state, touch sensors, and target info."""
        obs = collections.OrderedDict()
        obs["position"] = physics.bounded_position()
        obs["velocity"] = physics.velocity()
        obs["touch"] = physics.touch()
        obs["target_position"] = physics.target_position()
        obs["dist_to_target"] = physics.dist_to_target()
        return obs

    def get_reward(self, physics):
        return float(physics.dist_to_target() <= 0)


def _set_random_joint_angles(physics, random, max_attempts=1000):
    """Sets the joints to a random collision-free state."""

    for _ in range(max_attempts):
        randomizers.randomize_limited_and_rotational_joints(physics, random)
        # Check for collisions.
        physics.after_reset()
        if physics.data.ncon == 0:
            break
    else:
        raise RuntimeError(
            "Could not find a collision-free state "
            "after {} attempts".format(max_attempts)
        )



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved
"""Wrapper that samples a new task everytime the environment is reset."""

from mtenv import MTEnv
from mtenv.utils.types import ObsType
from mtenv.wrappers.multitask import MultiTask


class SampleRandomTask(MultiTask):
    def __init__(self, env: MTEnv):
        """Wrapper that samples a new task everytime the environment is
        reset.

        Args:
            env (MTEnv): Multitask environment to wrap over.
        """

        super().__init__(env=env)

    def reset(self) -> ObsType:
        self.env.reset_task_state()
        return self.env.reset()



# Copyright 2017 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

"""Acrobot domain."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import collections

from dm_control import mujoco
from dm_control.rl import control
from . import base
from . import common
from dm_control.utils import containers
from dm_control.utils import rewards
import numpy as np

_DEFAULT_TIME_LIMIT = 10
SUITE = containers.TaggedTasks()


def get_model_and_assets():
    """Returns a tuple containing the model XML string and a dict of assets."""
    return common.read_model("acrobot.xml"), common.ASSETS


@SUITE.add("benchmarking")
def swingup(time_limit=_DEFAULT_TIME_LIMIT, random=None, environment_kwargs=None):
    """Returns Acrobot balance task."""
    physics = Physics.from_xml_string(*get_model_and_assets())
    task = Balance(sparse=False, random=random)
    environment_kwargs = environment_kwargs or {}
    return control.Environment(
        physics, task, time_limit=time_limit, **environment_kwargs
    )


@SUITE.add("benchmarking")
def swingup_sparse(
    time_limit=_DEFAULT_TIME_LIMIT, random=None, environment_kwargs=None
):
    """Returns Acrobot sparse balance."""
    physics = Physics.from_xml_string(*get_model_and_assets())
    task = Balance(sparse=True, random=random)
    environment_kwargs = environment_kwargs or {}
    return control.Environment(
        physics, task, time_limit=time_limit, **environment_kwargs
    )


class Physics(mujoco.Physics):
    """Physics simulation with additional features for the Acrobot domain."""

    def horizontal(self):
        """Returns horizontal (x) component of body frame z-axes."""
        return self.named.data.xmat[["upper_arm", "lower_arm"], "xz"]

    def vertical(self):
        """Returns vertical (z) component of body frame z-axes."""
        return self.named.data.xmat[["upper_arm", "lower_arm"], "zz"]

    def to_target(self):
        """Returns the distance from the tip to the target."""
        tip_to_target = (
            self.named.data.site_xpos["target"] - self.named.data.site_xpos["tip"]
        )
        return np.linalg.norm(tip_to_target)

    def orientations(self):
        """Returns the sines and cosines of the pole angles."""
        return np.concatenate((self.horizontal(), self.vertical()))


class Balance(base.Task):
    """An Acrobot `Task` to swing up and balance the pole."""

    def __init__(self, sparse, random=None):
        """Initializes an instance of `Balance`.

        Args:
          sparse: A `bool` specifying whether to use a sparse (indicator) reward.
          random: Optional, either a `numpy.random.RandomState` instance, an
            integer seed for creating a new `RandomState`, or None to select a seed
            automatically (default).
        """
        self._sparse = sparse
        super(Balance, self).__init__(random=random)

    def initialize_episode(self, physics):
        """Sets the state of the environment at the start of each episode.

        Shoulder and elbow are set to a random position between [-pi, pi).

        Args:
          physics: An instance of `Physics`.
        """
        physics.named.data.qpos[["shoulder", "elbow"]] = self.random.uniform(
            -np.pi, np.pi, 2
        )
        super(Balance, self).initialize_episode(physics)

    def get_observation(self, physics):
        """Returns an observation of pole orientation and angular velocities."""
        obs = collections.OrderedDict()
        obs["orientations"] = physics.orientations()
        obs["velocity"] = physics.velocity()
        return obs

    def _get_reward(self, physics, sparse):
        target_radius = physics.named.model.site_size["target", 0]
        return rewards.tolerance(
            physics.to_target(), bounds=(0, target_radius), margin=0 if sparse else 1
        )

    def get_reward(self, physics):
        """Returns a sparse or a smooth reward, as specified in the constructor."""
        return self._get_reward(physics, sparse=self._sparse)
